import glob
from importlib.machinery import SourceFileLoader
import os
import sys
from types import ModuleType

from jolt import inspection
from jolt.tasks import attributes
from jolt.tasks import Alias, Task, TaskGenerator, TaskRegistry, WorkspaceResource
from jolt.error import raise_error_if, raise_task_error_if
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.manifest import ManifestExtension
from jolt.manifest import ManifestExtensionRegistry


class Recipe(object):
    def __init__(self, path, joltdir=None, project=None, source=None):
        self.path = path
        self.basepath = os.path.basename(path)
        self.joltdir = joltdir
        self.project = project
        self.source = source
        self.tasks = []

    def load(self):
        raise_error_if(self.source is not None, "recipe already loaded: {}", self.path)

        with open(self.path) as f:
            self.source = f.read()

    def save(self):
        raise_error_if(self.source is None, "recipe source unknown: {}", self.path)

        with open(self.path, "w") as f:
            f.write(self.source)


class NativeRecipe(Recipe):
    @staticmethod
    def _is_abstract(cls):
        return cls.__dict__.get("abstract", False) or cls.__name__.startswith("_")

    @staticmethod
    def _is_task(cls):
        return isinstance(cls, type) and \
            issubclass(cls, Task) and \
            not NativeRecipe._is_abstract(cls)

    def load(self):
        super(NativeRecipe, self).load()

        name = utils.canonical(self.path)
        loader = SourceFileLoader("joltfile_{0}".format(name), self.path)
        module = ModuleType(loader.name)
        module.__file__ = self.path
        loader.exec_module(module)
        sys.modules[loader.name] = module

        classes = inspection.getmoduleclasses(module, [Task, TaskGenerator], NativeRecipe._is_abstract)
        generators = []

        for cls in classes[TaskGenerator]:
            cls.joltdir = os.path.normpath(self.joltdir or os.path.dirname(self.path))
            generators.append(cls())

        for generator in generators:
            generated_tasks = utils.as_list(generator.generate())
            classes[Task] += filter(NativeRecipe._is_task, generated_tasks)

        for task in classes[Task]:
            task.name = task.name or task.__name__.lower()
            task.joltdir = os.path.normpath(self.joltdir or os.path.dirname(self.path))
            task.joltproject = self.project
            self.tasks.append(task)

        log.verbose("Loaded: {0}", self.path)


class Loader(object):
    def recipes(self):
        pass


class LoaderFactory(object):
    def create(self):
        raise NotImplementedError()


class NativeLoader(Loader):
    def __init__(self, searchpath):
        self._recipes = []
        self._find_files(searchpath)

    def _find_files(self, searchpath):
        files = glob.glob(fs.path.join(searchpath, "*.jolt"))
        for filepath in files:
            recipe = NativeRecipe(filepath)
            self._recipes.append(recipe)

    @property
    def recipes(self):
        return self._recipes


_loaders = []


def register(factory):
    raise_error_if(not issubclass(factory, LoaderFactory),
                   "{} is not a LoaderFactory", factory.__name__)
    _loaders.append(factory)


@register
class NativeLoaderFactory(LoaderFactory):
    def create(self, searchpath):
        return NativeLoader(searchpath)


@utils.Singleton
class JoltLoader(object):
    filename = "*.jolt"

    def __init__(self):
        self._recipes = []
        self._tasks = []
        self._path = None
        self._project_modules = {}
        self._project_recipes = {}
        self._project_resources = {}

    def _add_project_module(self, project, src):
        modules = self._project_modules.get(project, [])
        modules.append(src)
        self._project_modules[project] = modules

    def _get_project_modules(self, project):
        return self._project_modules.get(project, [])

    def _add_project_recipe(self, project, joltdir, src):
        recipes = self._project_recipes.get(project, [])
        recipes.append((joltdir, src))
        self._project_recipes[project] = recipes

    def _get_project_recipes(self, project):
        return self._project_recipes.get(project, [])

    def _add_project_resource(self, project, resource_name, resource_task):
        class ProjectResource(Alias):
            name = project + "/" + resource_name
            requires = [resource_task]

        self._tasks.append(ProjectResource)
        resources = self._project_resources.get(project, [])
        resources.append((resource_name, resource_task))
        self._project_resources[project] = resources

    def _get_project_resources(self, project):
        return self._project_resources.get(project, [])

    def _load_project_recipes(self):
        for project, recipes in self._project_recipes.items():
            resources = [resource for _, resource in self._get_project_resources(project)]
            for joltdir, src in recipes:
                joltdir = fs.path.join(self.joltdir, joltdir) if joltdir else self.joltdir
                recipe = NativeRecipe(fs.path.join(self._path, src), joltdir, project)
                recipe.load()
                for task in recipe.tasks:
                    task._resources = resources
                    attributes.requires("_resources")(task)
                self._tasks += recipe.tasks

    def _find_joltdir(self, searchdir):
        for factory in _loaders:
            loader = factory().create(searchdir)
            if loader.recipes:
                return searchdir

        parentdir = os.path.dirname(searchdir)
        if searchdir == parentdir:
            return None

        return self._find_joltdir(parentdir)

    def _get_searchpaths(self):
        return [self.joltdir]

    def load(self, manifest=None):
        if not self.joltdir:
            self.set_joltdir(self._find_joltdir(os.getcwd()))

        if not self.joltdir:
            return []

        for searchpath in self._get_searchpaths():
            for factory in _loaders:
                loader = factory().create(searchpath)
                for recipe in loader.recipes:
                    recipe.load()
                    self._recipes.append(recipe)
                    self._tasks += recipe.tasks

        self._load_project_recipes()
        return self._tasks

    def load_plugin(self, filepath):
        plugin, ext = os.path.splitext(fs.path.basename(filepath))
        loader = SourceFileLoader("jolt.plugins." + plugin, filepath)
        module = ModuleType(loader.name)
        module.__file__ = filepath
        loader.exec_module(module)
        sys.modules[loader.name] = module

    def load_plugins(self):
        searchpath = config.get("jolt", "pluginpath")
        searchpath = searchpath.split(":") if searchpath else []
        searchpath.append(fs.path.join(fs.path.dirname(__file__), "plugins"))

        for plugin in config.plugins():
            for path in searchpath:
                if "jolt.plugins." + plugin not in sys.modules:
                    module = fs.path.join(fs.path.dirname(__file__), path, plugin + ".py")
                    if fs.path.exists(module):
                        self.load_plugin(module)
                        continue

    @property
    def recipes(self):
        return self._recipes

    @property
    def tasks(self):
        return self._tasks

    @property
    def joltdir(self):
        return self._path

    def set_joltdir(self, value):
        if not self._path or len(value) < len(self._path):
            self._path = os.path.normpath(value) if value is not None else None


class RecipeExtension(ManifestExtension):
    def export_manifest(self, manifest, task):
        loader = JoltLoader.get()

        for recipe in loader.recipes:
            manifest_recipe = manifest.create_recipe()
            manifest_recipe.path = recipe.basepath
            manifest_recipe.source = recipe.source

        projects = set([task.task.joltproject for task in [task] + task.extensions + task.descendants])
        for project in filter(lambda x: x is not None, projects):
            manifest_project = manifest.create_project()
            manifest_project.name = project

            for name, resource_task in loader._get_project_resources(project):
                resource = manifest_project.create_resource()
                resource.name = name
                resource.text = resource_task

            for joltdir, src in loader._get_project_recipes(project):
                recipe = manifest_project.create_recipe()
                recipe.src = src
                if joltdir:
                    recipe.joltdir = joltdir

            for src in loader._get_project_modules(project):
                module = manifest_project.create_module()
                module.src = src

    def import_manifest(self, manifest):
        loader = JoltLoader.get()
        loader.set_joltdir(manifest.joltdir)

        for recipe in manifest.recipes:
            recipe = Recipe(recipe.path, source=recipe.source)
            recipe.save()

        for project in manifest.projects:
            for recipe in project.recipes:
                loader._add_project_recipe(project.name, recipe.joltdir, recipe.src)

            for resource in project.resources:
                loader._add_project_resource(project.name, resource.name, resource.text)

                # Acquire resource immediately
                task = TaskRegistry.get().get_task(resource.text, manifest=manifest)
                raise_task_error_if(not isinstance(task, WorkspaceResource), task,
                                    "only workspace resources are allowed in manifest")
                task.acquire_ws()

            for module in project.modules:
                loader._add_project_module(project.name, module.src)
                sys.path.append(fs.path.join(manifest.joltdir, module.src))


ManifestExtensionRegistry.add(RecipeExtension())


def get_workspacedir():
    workspacedir = JoltLoader.get().joltdir
    assert workspacedir is not None, "No workspace present"
    return workspacedir



from jolt import attributes as jolt_attributes
from jolt import utils
from jolt import Task


from jolt.pkgs import nodejs


class attributes(object):
    @staticmethod
    def flags(attrib):
        """
        Decorates a task with an alternative ``flags`` attribute.

        The new attribute will be concatenated with the regular
        ``flags`` attribute.

        Args:
            attrib (str): Name of alternative attribute.
                Keywords are expanded.
        """
        return utils.concat_attributes("flags", attrib)


@attributes.flags("flags")
@jolt_attributes.requires("requires_node")
class RunScript(Task):
    """ Runs an npm script """

    abstract = True

    flags = []
    """ List of build flags, passed directly to 'go build' """

    script = "build"
    """ Script name as defined in package.json """

    requires_node = ["nodejs"]
    """
    NodeJS requirement.

    Override to select another version or implementation.

    Example:

      .. code-block:: python

        requires_go = ["nodejs:version=16.15.1"]

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.script = self.expand(self.script)
        nodejs  # Dummy reference

    def run(self, deps, tools):
        self.info("Running {script}")
        tools.run("npm run {script} -- {}", " ".join(self._flags()))


class RunBuildScript(RunScript):
    """ Runs the npm build script and publishes the result """

    abstract = True

    script = "build"

    def run(self, deps, tools):
        self.outdir = tools.builddir()
        with tools.environ(BUILD_PATH=self.outdir):
            self.info("Installing deps")
            tools.run("npm ci --legacy-peer-deps")
            super().run(deps, tools)

    def publish(self, artifact, tools):
        with tools.cwd(self.outdir):
            artifact.collect("*")



from jolt import Download, Parameter, Resource, Task
from jolt.error import raise_task_error_if
from jolt.tasks import TaskRegistry
from jolt import attributes
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import tools
from jolt import utils

from jolt.cache import ArtifactListAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider

import contextlib
import json
from os import path
from platform import system
import tarfile


class DockerListVariable(ArtifactListAttribute):
    pass


class DockerLoadListVariable(DockerListVariable):
    def apply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for image in self.items():
            task.tools.run(
                "docker load -i {}",
                fs.path.join(artifact.path, image), output_on_error=True)


class DockerPullListVariable(DockerListVariable):
    def apply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for image in self.items():
            task.tools.run("docker pull {}", image, output_on_error=True)


class DockerRmiListVariable(DockerListVariable):
    def unapply(self, task, artifact):
        if isinstance(task, Resource):
            return
        for image in self.items():
            task.tools.run("docker rmi -f {}", image, output_on_error=True)


class DockerAttributeSet(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(DockerAttributeSet, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if name == "pull":
            return DockerPullListVariable(self._artifact, "pull")
        if name == "load":
            return DockerLoadListVariable(self._artifact, "load")
        if name == "rmi":
            return DockerRmiListVariable(self._artifact, "rmi")
        assert False, "No such docker attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class DockerAttributeProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "docker", DockerAttributeSet(artifact))

    def parse(self, artifact, content):
        if "docker" not in content:
            return
        for key, value in content["docker"].items():
            getattr(artifact.docker, key).set_value(value, expand=False)

    def format(self, artifact, content):
        if "docker" not in content:
            content["docker"] = {}
        for key, attrib in artifact.docker.items():
            content["docker"][key] = attrib.get_value()

    def apply(self, task, artifact):
        artifact.docker.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.docker.unapply(task, artifact)


class DockerClient(Download):
    """ Task: Downloads and publishes the Docker command line client.

    The task will be automatically made available after importing
    ``jolt.plugins.docker``.
    """

    name = "docker/cli"
    """ Name of the task """

    arch = Parameter("x86_64", help="Host architecture")
    """ Host architecture [x86_64] """

    collect = ["docker/docker"]

    host = Parameter(system().lower(), help="Host operating system")
    """ Host operating system [autodetected] """

    url = "https://download.docker.com/{host}/static/stable/{arch}/docker-{version}.tgz"
    """ URL of binaries """

    version = Parameter("20.10.13", help="Docker version")
    """ Docker version [20.10.13] """

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append("docker")


@attributes.requires("_image")
class DockerContainer(Resource):
    """
    Resource: Starts and stops a Docker container.
    """

    arguments = []
    """ Container argument list """

    cap_adds = []
    """ A list of capabilities to add to the container """

    cap_drops = []
    """ A list of capabilities to remove from the container """

    chroot = False
    """ Use as chroot - resource consumers will execute all commands in container """

    entrypoint = None
    """ Container entrypoint """

    environment = []
    """ Environment variables """

    image = None
    """
    Image tag or Jolt task.

    If a Jolt task is specified, its artifact must export a
    metadata string named ``tag`` with the name of the image tag.
    """

    labels = []
    """ A list of container metadata labels """

    privileged = False
    """
    Start container with elevated privileges.
    """

    ports = []
    """
    A list of container ports to publish.

    Example:

    .. code-block:: python

        ports = [
            "80",
            "443:443",
        ]

    Alternatively, assign ``True`` to publish all exposed ports to random ports.
    """

    volumes = []
    """
    A list of volumes to mount.

    By default, the cache directory and ``joltdir`` are automatically
    mounted in the container. See :attr:`volumes_default`.
    """

    volumes_default = [
        "{joltdir}:{joltdir}",
        "{joltcachedir}:{joltcachedir}",
    ]
    """
    A list of default volumes to mount.

    By default, the cache directory and ``joltdir`` are automatically
    mounted in the container. Override to disable.
    """

    user = None
    """
    Username or UID.

    Defaults to the current user.
    """

    @property
    def _arguments(self):
        return " ".join(self.arguments)

    @property
    def _cap_adds(self):
        return " ".join([utils.option("--cap-add ", cap) for cap in self.cap_adds])

    @property
    def _cap_drops(self):
        return " ".join([utils.option("--cap-drop ", cap) for cap in self.cap_drops])

    @property
    def _entrypoint(self):
        return utils.option("--entrypoint ", self.entrypoint)

    @property
    def _environment(self):
        return " ".join([utils.option("-e ", self.tools.expand(env)) for env in self.environment])

    @property
    def _image(self):
        registry = TaskRegistry.get()
        tool = tools.Tools(self)
        if registry.get_task_class(tool.expand(self.image)):
            return [self.image]
        return []

    def _info(self, fmt, *args, **kwargs):
        """
        Log information about the task.
        """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.info(fmt, *args, **kwargs)

    @property
    def _labels(self):
        return " ".join([utils.option("-l ", self.tools.expand(label)) for label in self.labels])

    @property
    def _privileged(self):
        return "--privileged" if self.privileged else ""

    @property
    def _ports(self):
        if self.ports is True:
            return "-P"
        return " ".join([utils.option("-p ", self.tools.expand(port)) for port in self.ports])

    @property
    def _user(self):
        if self.user:
            return f"--user {self.user}"
        try:
            from os import getuid
            return "--user " + str(getuid())
        except ImportError:
            return ""

    @property
    def _volumes(self):
        return " ".join([utils.option("-v ", self.tools.expand(vol))
                         for vol in self.volumes_default + self.volumes])

    def acquire(self, artifact, deps, tools, owner):
        self.joltcachedir = config.get_cachedir()
        try:
            image = deps[self.image]
            image = str(image.strings.tag)
        except Exception:
            image = tools.expand(self.image)

        self._info(f"Creating container from image '{image}'")
        self.container = tools.run(
            "docker run -i -d {_cap_adds} {_cap_drops} {_entrypoint} {_labels} {_ports} {_privileged} {_user} {_environment} {_volumes} {image} {_arguments}",
            image=image, output_on_error=True)

        self._info("Created container '{container}'")
        info = tools.run("docker inspect {container}", output_on_error=True)
        artifact.container = self.container
        artifact.info = json.loads(info)[0]

        if self.chroot:
            self._context_stack = contextlib.ExitStack()
            self._context_stack.enter_context(
                owner.tools.runprefix(f"docker exec -i {artifact.container}"))

    def release(self, artifact, deps, tools, owner):
        if self.chroot:
            self._context_stack.close()

        self._info("Stopping container '{container}'")
        tools.run("docker stop {container}", output_on_error=True)

        self._info("Deleting container '{container}'")
        tools.run("docker rm {container}", output_on_error=True)


class DockerLogin(Resource):
    """
    Resource: Logs in and out of a Docker Registry.

    If the user and password parameters are unset, credentials
    are fetched from the environment variables:

        - DOCKER_USER
        - DOCKER_PASSWD

    The resource will be automatically made available after importing
    ``jolt.plugins.docker``.
    """
    name = "docker/login"
    """ Name of the resource """

    requires = ["docker/cli"]

    user = Parameter("", help="Docker Registry username")
    """
    Docker Registry username.

    If not set, the environment variable ``DOCKER_USER`` is read instead.
    """

    passwd = Parameter("", help="Docker Registry password")
    """
    Docker Registry password.

    If not set, the environment variable ``DOCKER_PASSWD`` is read instead.
    """

    server = Parameter("", help="Docker Registry server")
    """
    Docker Registry server.

    If no server is specified, the default is defined by the daemon.
    """

    def _user(self, tools):
        return str(self.user) or tools.getenv("DOCKER_USER")

    def _password(self, tools):
        return str(self.passwd) or tools.getenv("DOCKER_PASSWD")

    def acquire(self, artifact, deps, tools, owner):
        raise_task_error_if(not self._user(tools), self, "Username has not been configured")
        raise_task_error_if(not self._password(tools), self, "Password has not been configured")

        with tools.cwd(tools.builddir()):
            tools.write_file("docker-credential", self._password(tools))
            tools.run("cat docker-credential | docker login -u {user} --password-stdin {server}", user=self._user(tools))

    def release(self, artifact, deps, tools, owner):
        tools.run("docker logout {server}")


TaskRegistry.get().add_task_class(DockerClient)
TaskRegistry.get().add_task_class(DockerLogin)


class _Tarfile(tarfile.TarFile):

    def without(self, targetpath):
        dirname, filename = fs.path.split(targetpath)
        if not filename.startswith(".wh."):
            return None
        fs.unlink(fs.path.join(dirname, filename[4:]), ignore_errors=True, tree=True)
        return True

    def makedev(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makedev(tarinfo, targetpath)

    def makedir(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makedir(tarinfo, targetpath)

    def makefifo(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makefifo(tarinfo, targetpath)

    def makefile(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        if fs.path.lexists(targetpath):
            fs.unlink(targetpath)
        super().makefile(tarinfo, targetpath)

    def makelink(self, tarinfo, targetpath):
        if self.without(targetpath):
            return
        super().makelink(tarinfo, targetpath)


class DockerImage(Task):
    """
    Abstract Task: Builds and publishes a Docker image.

    Builds the selected ``Dockerfile`` and optionally tags and pushes the
    image to a registry. The image can also be saved to file and published
    in the task artifact. Compression formats supported are bzip2, gzip and
    lzma.

    By default, base images referenced in the ``Dockerfile`` will be pulled
    during the build. Note that Jolt has no way of knowing beforehand if
    images have been updated in the registry. Use time-based influence to
    trigger rebuilds if it's important that base images are kept up-to-date.

    No automatic influence for ``Dockerfile`` or context is collected. Make
    sure to use an appropriate influence decorator.

    Optionally add requirements to:

      - ``docker/cli`` to provision the Docker client, if none is available on the host.
      - ``docker/login`` to automatically login to the Docker registry.

    This class must be subclassed.

    Example:

    .. code-block:: docker

        # Dockerfile

        FROM busybox:latest
        CMD ["busybox"]

    .. code-block:: python

        # build.jolt

        from jolt.plugins.docker import DockerImage

        class Busybox(DockerImage):
            \"\"\" Publishes Busybox image as gzip-compressed tarball \"\"\"
            compression = "gz"
            requires = ["docker/cli"]
            tags = ["busybox:{identity}"]

    """
    abstract = True

    autoload = True
    """
    Automatically load image file into local registry when the artifact is
    consumed by another task.

    If the built image is saved to a file (i.e. ``imagefile`` is set), the image
    file is automatically loaded into the local Docker registry when the task
    artifact is consumed by another task. The image is also automatically
    removed from the registry upon completion of the consumer task.

    Default: ``True``.
    """

    buildargs = []
    """
    List of build arguments and their values ("ARG=VALUE").

    The arguments are passed to Docker using ``--build-arg``.
    """

    cleanup = True
    """ Remove image from Docker daemon upon completion [True] """

    compression = None
    """ Optional image compression "bz2", "gz", or "xz". """

    context = "."
    """ Path to build context, relative to joltdir (directory). """

    dockerfile = "Dockerfile"
    """ Path to the Dockerfile to build, or the full source code of such a file. """

    extract = False
    """
    Extract image and publish rootfs tree.

    This option is useful when building a chroot to be used with
    :func:`jolt.Tools.chroot`. It disables saving of the image to a tarball.
    """

    imagefile = "{canonical_name}.tar"
    """
    Name of the image tarball published by the task.

    If set to None, no image file will be saved and published.

    Defaults to the task's canonical name.
    """

    labels = []
    """ A list of image metadata labels """

    platform = None
    """ Target platform, e.g. linux/arm/v7. """

    pull = True
    """
    Always pull images when building.

    Passes --pull to the Docker client.
    """

    push = False
    """
    Optionally push image to registry [False]

    To be able to push images, the current user must login to the Docker Registry.
    The ``docker/login`` Jolt resource can be used for that purpose.
    """

    squash = False
    """ Squash image layers """

    tags = ["{canonical_name}:{identity}"]
    """ Optional list of image tags. Defaults to task's canonical name. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _extract_layer(self, tools, layerpath, targetpath):
        layerpath = tools.expand_path(layerpath)
        targetpath = tools.expand_path(targetpath)

        with _Tarfile.open(layerpath, 'r') as tar:
            tar.extractall(targetpath)

    @property
    def _buildargs(self):
        return " ".join([utils.option("--build-arg ", self.tools.expand(ba)) for ba in self.buildargs])

    @property
    def _labels(self):
        return " ".join([utils.option("-l ", self.tools.expand(label)) for label in self.labels])

    @property
    def _platform(self):
        platform = self.tools.expand(self.platform) if self.platform else None
        return utils.option("--platform ", platform)

    @property
    def _tags(self):
        return " ".join([utils.option("-t ", tag) for tag in self.tags])

    def run(self, deps, tools):
        context = tools.expand_relpath(self.context, self.joltdir)
        dockerfile = tools.expand_path(self.dockerfile)
        self._imagefile = tools.expand(self.imagefile) if self.imagefile else None
        self._autoload = self._imagefile and self.autoload
        self.tags = [self.tools.expand(tag) for tag in self.tags]
        pull = " --pull" if self.pull else ""
        squash = " --squash" if self.squash else ""

        # If dockerfile is not relative to joltdir, look for it in context
        if not path.exists(dockerfile):
            with tools.cwd(context):
                dockerfile = tools.expand_path(self.dockerfile)

        if not path.exists(dockerfile):
            with tools.cwd(tools.builddir()):
                tools.write_file("Dockerfile", self.dockerfile)
                dockerfile = tools.expand_path("Dockerfile")

        self.info("Building image from {} in {}",
                  tools.expand_relpath(dockerfile),
                  tools.expand_relpath(context))

        with tools.cwd(context):
            tools.run("docker build {_platform} . -f {} {_buildargs} {_labels} {_tags} {pull}{squash}",
                      utils.quote(dockerfile), pull=pull, squash=squash)

        try:
            if self.push:
                self.info("Pushing image")
                for tag in self.tags:
                    tools.run("docker push {}", tag)

            if self._imagefile or self.extract:
                self.info("Saving image to file")
                with tools.cwd(tools.builddir()):
                    tools.run("docker image save {} -o {}", self.tags[0], self._imagefile or "image.tar")

            if self.extract:
                with tools.cwd(tools.builddir()):
                    tools.extract(self._imagefile or "image.tar", "layers/")
                    manifest = json.loads(tools.read_file("layers/manifest.json"))
                    for image in manifest:
                        for layer in image.get("Layers", []):
                            self.info("Extracting layer {}", fs.path.dirname(layer))
                            self._extract_layer(tools, fs.path.join("layers", layer), "rootfs/")
            elif self._imagefile:
                with tools.cwd(tools.builddir()):
                    if self.compression is not None:
                        tools.compress("{_imagefile}", "{_imagefile}.{compression}")

        finally:
            if self.cleanup:
                self.info("Removing image from Docker daemon")
                for tag in self.tags:
                    utils.call_and_catch(tools.run("docker image rm {}", tag))

    def publish(self, artifact, tools):
        artifact.strings.tag = tools.expand(self.tags[0])
        if self.extract:
            with tools.cwd(tools.builddir()):
                artifact.collect("rootfs", symlinks=True)
            artifact.paths.rootfs = "rootfs"
        elif self._imagefile:
            with tools.cwd(tools.builddir()):
                if self.compression is not None:
                    artifact.collect("{_imagefile}.{compression}")
                    if self._autoload:
                        artifact.docker.load.append("{_imagefile}.{compression}")
                else:
                    artifact.collect("{_imagefile}")
                    if self._autoload:
                        artifact.docker.load.append("{_imagefile}")
            if self._autoload:
                artifact.docker.rmi.append(artifact.strings.tag.get_value())



from contextlib import contextmanager
import copy
import smtplib
from email.message import EmailMessage
import os
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.error import raise_error_if
from jolt.plugins import report
from jolt.hooks import CliHook, CliHookFactory
from jolt.tasks import TaskRegistry


class EmailHooks(CliHook):
    def __init__(self):
        self._server = config.get("email", "server")
        self._to = config.get("email", "to")
        self._cc = config.get("email", "cc")
        self._bcc = config.get("email", "bcc")
        self._from = config.get("email", "from", "jolt@localhost")
        self._subject = config.get("email", "subject", "Jolt Build Report")
        self._stylesheet = config.get(
            "email", "stylesheet",
            fs.path.join(fs.path.dirname(__file__), "email.xslt"))
        self._artifact = config.get("email", "artifact")
        self._failure = config.getboolean("email", "on_failure", True)
        self._success = config.getboolean("email", "on_success", True)
        raise_error_if(not self._server, "email.server not configured")
        raise_error_if(not self._server, "email.to not configured")

    def annotate_report(self, report):
        # Make environment variables available to stylesheet
        for key, value in os.environ.items():
            param = report.create_parameter()
            param.key = key
            param.value = value

    def shorten_task_names(self, report):
        # Use the short qualified task names in the report
        for mtask in report.tasks:
            with utils.ignore_exception():
                # mtask.name is a fully qualified name
                task = TaskRegistry.get().get_task(mtask.name)
                mtask.name = task.short_qualified_name

    def send_report(self, report):
        self.annotate_report(report)
        self.shorten_task_names(report)

        if self._artifact:
            dirname = fs.path.dirname(self._artifact)
            if dirname:
                fs.makedirs(dirname)
            with open(self._artifact, "w") as f:
                f.write(report.transform(self._stylesheet))

        if report.has_failure():
            if not self._failure:
                return
        else:
            if not self._success:
                return

        msg = EmailMessage()
        msg['Subject'] = self._subject
        msg['From'] = self._from
        msg['To'] = ", ".join(utils.unique_list(self._to.split()))
        if self._cc:
            msg['Cc'] = ", ".join(utils.unique_list(self._cc.split()))
        if self._bcc:
            msg['Bcc'] = ", ".join(utils.unique_list(self._bcc.split()))
        msg.set_content("Your e-mail client cannot display HTML formatted e-mails.")
        msg.add_alternative(report.transform(self._stylesheet), subtype='html')

        with smtplib.SMTP(self._server) as server:
            log.info("Sending email report to {}", self._to)
            server.send_message(msg)

    @contextmanager
    def cli_build(self, *args, **kwargs):
        start = utils.duration()
        try:
            yield
        finally:
            with report.update() as manifest:
                manifest = copy.copy(manifest)
                manifest.duration = str(int(start.seconds))
                self.send_report(manifest)


@CliHookFactory.register
class ResultFactory(CliHookFactory):
    def create(self, env):
        return EmailHooks()



from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider
from jolt.cache import ArtifactStringAttribute

from jolt import filesystem as fs


class PathVariable(ArtifactStringAttribute):
    def __init__(self, artifact, name):
        super(PathVariable, self).__init__(artifact, name)
        self._old_value = None

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass

    def __str__(self):
        if self._value is None:
            return ""
        return fs.path.join(self._artifact.path, self._value)


class PathVariableSet(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(PathVariableSet, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        return PathVariable(self._artifact, name)


@ArtifactAttributeSetProvider.Register
class PathVariableSetProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "paths", PathVariableSet(artifact))

    def parse(self, artifact, content):
        if "paths" not in content:
            return

        for key, value in content["paths"].items():
            getattr(artifact.paths, key).set_value(value, expand=False)

    def format(self, artifact, content):
        if "paths" not in content:
            content["paths"] = {}

        for key, value in artifact.paths.items():
            content["paths"][key] = value.get_value()

    def apply(self, task, artifact):
        artifact.paths.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.paths.unapply(task, artifact)



from .tasks import Alias
from .tasks import BooleanParameter
from .tasks import Chroot
from .tasks import Download
from .tasks import EnvironExport
from .tasks import Export
from .tasks import ListParameter
from .tasks import MultiTask
from .tasks import Parameter
from .tasks import Resource
from .tasks import Script
from .tasks import Task
from .tasks import TaskGenerator
from .tasks import Test
from .tasks import attributes

from .cache import Artifact
from .cache import Context

from .tools import Tools

from .version import __version__
from . import expires
from . import influence

__all__ = (
    "Alias",
    "Artifact",
    "BooleanParameter",
    "Chroot",
    "Context",
    "Download",
    "EnvironExport",
    "Export",
    "ListParameter",
    "MultiTask",
    "Parameter",
    "Resource",
    "Script",
    "Task",
    "TaskGenerator",
    "Test",
    "Tools",
    "__version__",
    "attributes",
    "expires",
    "influence",
)

name = "jolt"


def include(joltfile):
    """ Include another Python file """
    try:
        from os import path
        from sys import _getframe
        from jolt.loader import JoltLoader
        filepath = _getframe().f_back.f_code.co_filename
        filepath = path.dirname(filepath)
        filepath = path.join(filepath, joltfile)
        JoltLoader.get()._load_file(filepath)
    except Exception as e:
        from jolt.error import raise_error
        raise_error("Failed to load '{0}': {1}", joltfile, str(e))



from xml.etree import ElementTree as ET


def _tcname(suite, case):
    return "{}.{}".format(suite.get("name"), case.get("name"))


def import_junit_report(xml, report, errors=True, failures=True):
    """
    Import errors and failures from a JUnit XML report.

    Args:
        xmlreport (str): JUnit report data to be imported (XML string).
        report: a Jolt task report, returned by ``Task.report()``
        errors (boolean): import errors from the JUnit report
        failures (boolean): import failures from the JUnit report
    """

    tree = ET.parse(xml)
    root = tree.getroot()
    testsuites = root.findall(".//testsuite")
    for testsuite in testsuites:
        testcases = testsuite.findall(".//testcase")
        for testcase in testcases:
            if errors:
                tcerrors = testcase.findall(".//error")
                for error in tcerrors:
                    name = _tcname(testsuite, testcase)
                    message = error.text
                    report.add_error("Test Error", name, message)
            if failures:
                tcfailures = testcase.findall(".//failure")
                for failure in tcfailures:
                    name = _tcname(testsuite, testcase)
                    message = failure.text
                    report.add_error("Test Failed", name, message)



import base64
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager, ExitStack
import copy
import fnmatch
import functools
import hashlib
import platform
import subprocess
from os import environ
import sys
import unittest as ut
import uuid
import re
import traceback

from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.cache import ArtifactAttributeSetProvider
from jolt.error import raise_error_if, raise_task_error, raise_task_error_if
from jolt.error import raise_unreported_task_error_if
from jolt.error import JoltError, JoltCommandError
from jolt.expires import Immediately
from jolt.influence import FileInfluence, TaintInfluenceProvider
from jolt.influence import TaskClassSourceInfluence
from jolt.influence import attribute as attribute_influence
from jolt.influence import environ as environ_influence
from jolt.influence import source as source_influence
from jolt.influence import files as file_influence
from jolt.manifest import _JoltTask
from jolt.tools import Tools
from jolt import colors


class Export(object):

    def __init__(self, value, encoded=False):
        self._imported = False
        self._task = None
        self._value = None
        self.exported_value = value
        self.encoded = encoded

    @staticmethod
    def __get_exports__(obj):
        exports = {}
        for mro in reversed(obj.__class__.__mro__):
            for key, export in getattr(mro, "__export_list", {}).items():
                attr = getattr(obj.__class__, key)
                if isinstance(attr, Export):
                    exports[key] = attr
        return exports

    def __set_name__(self, owner, name):
        if "__export_list" not in owner.__dict__:
            setattr(owner, "__export_list", {})
        getattr(owner, "__export_list")[name] = self

    @property
    def value(self):
        if self._value is None:
            self._value = self.exported_value(self._task)
        return self._value

    @property
    def is_imported(self):
        return self._imported

    def assign(self, value):
        value = value or ""
        self._value = base64.decodebytes(value.encode()).decode() if self.encoded else value
        self._imported = True

    def export(self, task):
        value = self.value
        if value:
            value = base64.encodebytes(value.encode()).decode() if self.encoded else value
        return value

    def set_task(self, task):
        self._task = task

    def __str__(self):
        return str(self.value)


class EnvironExport(Export):
    def __init__(self, envname):
        super().__init__(value=lambda self: environ.get(envname))
        self._envname = envname

    def assign(self, value):
        super().assign(value)
        self._task.tools.setenv(self._envname, value)


class Parameter(object):
    """ Generic task parameter type. """

    @staticmethod
    def __get_params__(obj):
        params = {}
        for mro in reversed(obj.__class__.__mro__):
            for key, param in getattr(mro, "__param_list", {}).items():
                attr = getattr(obj.__class__, key)
                if isinstance(attr, Parameter):
                    params[key] = attr
        return params

    def __set_name__(self, owner, name):
        if "__param_list" not in owner.__dict__:
            setattr(owner, "__param_list", {})
        getattr(owner, "__param_list")[name] = self

    def __init__(self, default=None, values=None, required=True,
                 const=False, influence=True, help=None):
        """
        Creates a new parameter.

        Args:
            default (str, optional): An optional default value.
            values (list, optional): A list of accepted values. An
                assertion is raised if an unlisted value is assigned to the parameter.
            required (boolean, optional): If required, the parameter must be assigned
                a value before the task can be executed. The default is ``True``.
            const (boolean, optional): If const is True, the parameter is immutable
                and cannot be assigned a non-default value. This is useful in
                a class hierarchy where a subclass may want to impose restrictions
                on a parent class parameter. The default is ``False``.
            influence (boolean, optional): If influence is False, the parameter value
                will not influence the identity of the task artifact. The default is
                True.
            help (str, optional): Documentation for the parameter.
                This text is displayed when running the ``info`` command on the
                associated task.

        Raises:
            ValueError: If the parameter is assigned an illegal value.

        """

        self._default = default
        self._value = default
        self._accepted_values = values
        self._required = required
        self._const = const
        self._influence = influence
        self._help = help
        if default:
            self._validate(default)

    @property
    def help(self):
        values = self._help_values()
        if values:
            return "{} {}".format(self._help, values) if self._help else values
        return self._help or ""

    def _help_values(self, accepted=None):
        accepted = accepted or self._accepted_values or []

        def highlight(value):
            return colors.bright(value) if self._is_default(value) else colors.dim(value)

        return "[{}]".format(", ".join([highlight(value) for value in accepted])) if accepted else ""

    def __str__(self):
        """ Returns the parameter value as a string """
        return str(self._value) if self._value is not None else ''

    def _validate(self, value):
        if self._accepted_values is not None and value not in self._accepted_values:
            raise ValueError(value)

    def get_default(self):
        """ Get the default value of the parameter.

        Returns:
            The default value or None if no default was given.
        """
        return self._default

    @property
    def default(self):
        """ Get the default value of the parameter.

        Returns:
            The default value or None if no default was given.
        """
        return self._default

    def set_default(self, value):
        """ Set the default value of the parameter.

        Args:
            value (str): The new default value of the parameter.
        """
        self._validate(value)
        if self.is_default():
            self._value = value
        self._default = value

    def is_influencer(self):
        """ Check if the parameter value influences the identitiy of the task artifact.

        Returns:
            True if the parameter influences the identity, False otherwise.
        """
        return self._influence

    def is_required(self):
        """ Check if the parameter must be set to a value.

        Returns:
            True if the parameter must be assigned a value, False otherwise.
        """
        return self._required

    def is_default(self):
        """ Check if parameter is set to its default value.

        Returns:
            True if the assigned value is the default value.
        """
        return self._default == self._value

    def _is_default(self, value):
        return self._default and value == self._default

    def is_unset(self):
        """ Check if the parameter is unset.

        Returns:
            True if the assigned value is None.
        """
        return self._value is None

    def is_set(self):
        """ Check if the parameter is set to a non-default value.

        Returns:
            True if the assigned value is not the default value.
        """
        return not self.is_unset() and not self.is_default()

    def is_const(self):
        """ Check if the parameter can be assigned or not.

        Returns:
            True if the parameter is immutable.
        """
        return self._const

    def get_value(self):
        """ Get the parameter value. """
        return self._value

    @property
    def value(self):
        """ Get the parameter value. """
        return self._value

    def __bool__(self):
        """ Returns True if the parameter value is a non-empty string. """
        return self._value is not None and self._value != ""

    def __eq__(self, value):
        """ Compare parameter value """
        return self._value == value

    def set_value(self, value):
        """ Set the parameter value.

        Args:
            value (str): The new parameter value.

        Raises:
            ValueError: If the parameter is assigned an illegal value.
        """
        self._validate(value)
        if self._const and value != self._default:
            raise ValueError("immutable")
        self._value = value


class BooleanParameter(Parameter):
    """
    Boolean task parameter.

    Accepted values are:

      - False
      - True
      - "false"
      - "true"
      - "no"
      - "yes"
      - "0"
      - "1"

    """

    def __init__(self, default=None, required=True, const=False, influence=True, help=None):
        """
        Creates a new parameter.

        Args:
            default (boolean, optional): An optional default boolean value.
            required (boolean, optional): If required, the parameter must be assigned
                a value before the task can be executed. The default is ``True``.
            const (boolean, optional): If const is True, the parameter is immutable
                and cannot be assigned a non-default value. This is useful in
                a class hierarchy where a subclass may want to impose restrictions
                on a parent class parameter. The default is ``False``.
            influence (boolean, optional): If influence is False, the parameter value
                will not influence the identity of the task artifact. The default is
                True.
            help (str, optional): Documentation for the parameter.
                This text is displayed when running the ``info`` command on the
                associated task.

        Raises:
            ValueError: If the parameter is assigned an illegal value.

        """
        default = str(default).lower() if default is not None else None
        super(BooleanParameter, self).__init__(
            default,
            values=["false", "true", "0", "1", "no", "yes"],
            required=required,
            const=const,
            influence=influence,
            help=help)

    def _help_values(self, accepted=None):
        return super()._help_values(["true", "false"])

    def set_value(self, value):
        """ Set the parameter value.

        Args:
            value (boolean): The new parameter value. Accepted values are:
                False, True, "false, and "true", 0 and 1, "no" and "yes".

        Raises:
            ValueError: If the parameter is assigned an illegal value.
        """
        value = str(value).lower()
        super(BooleanParameter, self).set_value(value)

    @property
    def is_true(self):
        """ The parameter value is True. """
        return str(self.get_value()) in ["true", "1", "yes"]

    @property
    def is_false(self):
        """ The parameter value is False. """
        return not self.is_true

    def __bool__(self):
        """ Returns the boolean parameter value """
        return self.is_true

    def __getitem__(self, key):
        """
        Returns a substitution string depending on the parameter value.

        Args:
            key (str): A special key syntax, ``enabled,disabled``, where
            and either ``enabled`` or ``disabled`` would be returned
            depending on the paramter value. See the example below.

        Returns:
            Substitution string.

        Example:

          .. code-block:: python

            class Example(Task):
               debug = BooleanParameter()

               def run(self, deps, tools):
                   self.info("debug is {debug[enabled,disabled]}")

          .. code-block:: bash

            $ jolt build example:debug=true
            [INFO] debug is enabled (example)

            $ jolt build example:debug=false
            [INFO] debug is disabled (example)

        """
        key = key.split(",")
        if len(key) != 2:
            raise KeyError(key)
        return key[0] if self.is_true else key[1]


class ListParameter(Parameter):
    """ List parameter type.

    A list parameter allows multiple values to be assigned to it. Values are
    separated by the '+' character in qualified task names. Each assigned value
    is validated against the list of accepted values. They are sorted in
    alphabetical order before the task is executed.

    Example:

      .. code-block:: python

        class Example(Task):
            arg = ListParameter(default=["c"], values=["a", "b", "c"], help="A list parameter")

            def run(self, deps, tools):
                for item in self.arg:
                    print(item)

      .. code-block:: bash

        $ jolt build example example:arg=a example:arg=a+b

    """

    def __init__(self, *args, **kwargs):
        """
        Creates a new list parameter.

        Args:
            default (boolean, optional): An optional list of default values.
            values (list, optional): A list of accepted values. An
                assertion is raised if an unlisted value is assigned to the parameter.
            required (boolean, optional): If required, the parameter must be assigned
                a value before the task can be executed. The default is ``True``.
            const (boolean, optional): If const is True, the parameter is immutable
                and cannot be assigned a non-default value. This is useful in
                a class hierarchy where a subclass may want to impose restrictions
                on a parent class parameter. The default is ``False``.
            influence (boolean, optional): If influence is False, the parameter value
                will not influence the identity of the task artifact. The default is
                True.
            help (str, optional): Documentation for the parameter.
                This text is displayed when running the ``info`` command on the
                associated task.

        Raises:
            ValueError: If the parameter is assigned an illegal value.

        """
        super().__init__(*args, **kwargs)

    def set_value(self, value):
        """ Set the parameter value.

        Args:
            value (str): A list of accepted values, separated by '+'.

        Raises:
            ValueError: If the parameter is assigned an illegal value.
        """
        value = str(value).split("+") if type(value) == str else value
        value.sort()
        super().set_value(value)

    def _validate(self, value):
        if self._accepted_values is not None:
            for item in value:
                if item not in self._accepted_values:
                    raise ValueError(item)

    def get_value(self):
        return "+".join(self._value)

    def _is_default(self, value):
        return self._default and value in self._default

    def __bool__(self):
        """ Returns True if the list is non-empty. """
        return len(self._value) > 0

    def __iter__(self):
        """ Returns a sequence iterator. """
        return iter(self._value)

    def __len__(self):
        """ Returns the length of the list. """
        return len(self._value)

    def __getitem__(self, key):
        """
        Returns an element or a slice from the list.

        Args:
            key (int, slice, str): Element index or slice. A key string can
                be used to check for the existence of that value in the list.
                If the key is present the same value is returned, otherwise None.

                A special key syntax is also available to request an alternate return
                value depending on the presence of the key. Instead of a list value
                you pass ``value,present,absent`` and either ``present`` or ``absent``
                will be returned. See the example below.

        Returns:
            Element value, or substitution.

        Example:

          .. code-block:: python

            class Example(Task):
               features = ListParameter(values=["optimize", "strip"], required=False)

               def run(self, deps, tools):
                   if len(self.features) > 0:
                       self.info("first feature is {features[0]}")
                       self.info("optimize == {features[optimize]}")
                       self.info("optimization is {features[optimize,enabled,disabled]}")

          .. code-block:: bash

            $ jolt build example:features=optimize+strip
            [INFO] first feature is optimize (example)
            [INFO] optimize = optimize (example)
            [INFO] optimization is enabled (example)

            $ jolt build example:features=strip
            [INFO] first feature is debug (example)
            [INFO] optimize = None (example)
            [INFO] optimization is disabled (example)

        """
        if type(key) == str:
            key = key.split(",")
            if len(key) == 1:
                true = key[0]
                false = None
            else:
                true = key[1] if len(key) > 1 else key[0]
                false = key[2] if len(key) > 2 else None
            return true if key[0] in self._value else false
        return self._value[key]


class TaskRegistry(object):
    _instance = None

    def __init__(self, env=None):
        self.env = env
        self.tasks = {}
        self.instances = {}

    @staticmethod
    def get(*args, **kwargs):
        if not TaskRegistry._instance:
            TaskRegistry._instance = TaskRegistry(*args, **kwargs)
        return TaskRegistry._instance

    def add_task_class(self, cls):
        self.tasks[cls.name] = cls

    def add_task(self, task, extra_params):
        name, params = utils.parse_task_name(task.name)
        params.update(extra_params or {})
        full_name = utils.format_task_name(name, params)
        self.instances[full_name] = task

    def get_task_class(self, name):
        return self.tasks.get(name)

    def get_task_classes(self):
        return list(self.tasks.values())

    def get_task(self, name, extra_params=None, manifest=None):
        name, params = utils.parse_task_name(name)
        params.update(extra_params or {})
        full_name = utils.format_task_name(name, params)

        task = self.instances.get(full_name)
        if task:
            return task

        cls = self.tasks.get(name)
        if cls:
            task = cls(parameters=params, manifest=manifest)
            task = self.instances.get(task.qualified_name, task)
            self.instances[task.qualified_name] = task
            self.instances[full_name] = task
            return task

        raise_task_error_if(not task, full_name, "No such task")

    def set_default_parameters(self, task):
        name, params = utils.parse_task_name(task)

        cls = self.tasks.get(name)
        raise_task_error_if(not cls, task, "No such task")
        cls._set_default_parameters(cls, params)

    def set_joltdir(self, joltdir):
        for task in list(self.tasks.values()):
            task.joltdir = joltdir

    def _create_parents(self, name):
        names = name.split("/")

        if len(names) <= 1:
            return

        prev = None
        for i in reversed(range(0, len(names))):
            name = "/".join(names[:i + 1])
            task = self.instances.get(name) or Task(name)
            if prev:
                task.requires += [prev.name]
            prev = task


class TaskGenerator(object):
    """
    Base class for task generators.

    Generators create tasks classes dynamically in runtime.
    They are used to implement concepts such as components
    where multiple similar tasks should be created.

    For example, a C++ library should not only be built, it
    should also be tested. While it is possible to add one
    task for the compilation of the library and another for
    the test, it is advisable to instead use a task generator
    when there are multiple libraries to build and test.
    The generator will ensure that all libraries are built and
    tested identically. Additional tasks, such as sourcedoc
    generation, can be added for all libraries at a later
    point in time. Refactoring is simplified by gathering
    shared code in one place.

    Generated tasks can be replaced by defining a task with
    the same name explicitly.

    """

    abstract = True
    """ An abstract task generator class indended to be subclassed.

    Abstract task generators are not invoked.
    """

    def generate(self):
        """
        Generate tasks.

        Called by Jolt during the parsing of user task
        definitions.

        Returns:
            A list of task classes to be registered with Jolt.

        """
        raise NotImplementedError()


class attributes:
    @staticmethod
    def requires(attrib):
        """
        Decorates a task with an alternative ``requires`` attribute.

        The new attribute will be concatenated with the regular
        ``requires`` attribute.

        Args:
            attrib (str): Name of alternative attribute.
                Keywords are expanded.
        """
        return utils.concat_attributes("requires", attrib)

    @staticmethod
    def attribute(alias, target, influence=True, default=False):
        """
        Decorates a task with an alias for another attribute.

        Args:
            attrib (str): Name of alias attribute.
            target (str): Name of target attribute.
                Keywords are expanded.
            influence (boolean): Add value of target
                attribute as influence of the task.
            default (boolean): Return alias attribute if
                target attribute does not exist. Value is
                accessed through the alias attribute name
                with a leading underscore, e.g. '_alias'.

        """
        def _decorate(cls):
            def _get(self):
                if default:
                    return getattr(self, self.expand(target), getattr(self, self.expand(alias)))
                return getattr(self, self.expand(target))
            if default:
                setattr(cls, "_" + alias, property(_get))
            else:
                setattr(cls, alias, property(_get))
            if influence:
                attribute_influence(target)(cls)
            return cls
        return _decorate

    @staticmethod
    def environ(envname, influence=True):
        """
        Decorator marking the task as dependent on an environment variable.

        The value of the environment variable will be automatically
        transferred to workers in distributed network builds.

        Args:
            envname (str): Name of the environment variable.
            influence (boolean): Add value of environment
                variable as influence of the task. Default: True.

        """
        def _decorate(cls):
            if influence:
                environ_influence(envname)(cls)
            name = "_environ_" + utils.canonical(envname.lower())
            export = EnvironExport(envname)
            export.__set_name__(cls, name)
            setattr(cls, name, export)
            return cls
        return _decorate

    @staticmethod
    def publish_files(attrib):
        """
        Decorator adding a list attribute where file publication can be specified.

        Each item in the list is a set of arguments passed directly to
        :func:`jolt.Artifact.collect`. Tuples, dictionaries and strings are
        accepted.

        Example:

        .. code-block:: python

          @jolt.attributes.publish_files("collect")
          class Example(Task):
              collect = [
                  # Publish file.txt into artifact root
                  "file.txt",

                  # Publish file.txt into artifact dir/ directory
                  ("file.txt", "dir/"),

                  # Publish files from dir/ into artifact root
                  {"files": "*", "cwd": "dir"},
              ]

        """

        def decorate(cls):
            if not hasattr(cls, "__publish_files"):
                old_pub = cls.publish

                def publish(self, artifact, tools):
                    old_pub(self, artifact, tools)
                    for args in getattr(self, "__publish_files")():
                        if type(args) == tuple:
                            artifact.collect(*args)
                        elif type(args) == dict:
                            artifact.collect(**args)
                        else:
                            artifact.collect(args)

                cls.publish = publish

            return utils.concat_attributes("_publish_files", attrib)(cls)
        return decorate

    @staticmethod
    def load(filepath):
        """
        Decorator which loads task class attributes from a file.

        The loaded file is Python source file declaring a dictionary
        with keys and values to be assigned to the task instance.

        The file is automatically registered as task hash influence.

        Example:

        .. code-block:: python

          @attributes.load("attributes-{os}.py")
          class Print(Task):
              os = Parameter()

              def run(self, deps, tools):
                  print("OS Author: ", self.os_author)

        .. code-block:: python

          # attributes-linux.py
          {
              "os_author": "Torvalds",
          }

        .. code-block:: bash

          $ jolt build print:os=linux

        """
        def decorate(cls):
            cls = file_influence(filepath)(cls)
            old_init = cls.__init__

            def new_init(self, *args, **kwargs):
                old_init(self, *args, **kwargs)
                for key, val in eval(self.tools.read_file(filepath)).items():
                    setattr(self, key, val)
                    self.requires = utils.unique_list(
                        utils.call_or_return_list(self, self.__class__._requires))
                self.requires = self.expand(self.requires)

            cls.__init__ = new_init
            return cls
        return decorate

    @staticmethod
    def method(alias, target, influence=True):
        """
        Decorates a task with an alias for another method.

        Args:
            attrib (str): Name of alias method.
            target (str): Name of target method.
                Keywords are expanded.
            influence (boolean): Add source of target
                method as influence of the task.

        """
        def _decorate(cls):
            def _call(self, *args, **kwargs):
                return getattr(self, self.expand(target))(*args, **kwargs)
            setattr(cls, alias, _call)
            if influence:
                source_influence(target)(cls)
            return cls
        return _decorate

    @staticmethod
    def system(cls):
        """
        Decorates a task with a property returning the operating system name.

        Examples: "linux", "windows"
        """
        cls._system = Export(lambda t: platform.system().lower())
        cls.system = property(lambda t: t._system.value)
        return cls


class TaskBase(object):
    """ Task base class. """

    cacheable = True
    """ Whether the task produces an artifact or not. """

    expires = Immediately()
    """An expiration strategy, defining when the artifact may be evicted from the cache.

    When the size of the artifact cache exceeds the configured limit
    an attempt will be made to evict artifacts from the cache. The eviction
    algorithm processes artifacts in least recently used (LRU) order until
    an expired artifact is found.

    By default, an artifact expires immediately and may be evicted at any time
    (in LRU order). An exception to this rule is if the artifact is required by
    a task in the active task set. For example, if a task A requires the output
    of task B, B will never be evicted by A while A is being executed.

    There are several expiration strategies to choose from:

     - :class:`jolt.expires.WhenUnusedFor`
     - :class:`jolt.expires.After`
     - :class:`jolt.expires.Never`

    Examples:

        .. code-block:: python

          # May be evicted if it hasn't been used for 15 days
          expires = WhenUnusedFor(days=15)

        .. code-block:: python

          # May be evicted 1h after creation
          expires = After(hours=1)

        .. code-block:: python

          # Never evicted
          expires = Never()

    """

    extends = ""
    """
    Name of extended task.

    A task with this attribute set is called an extension. An extension
    is executed in the context of the extended task, immediately after
    the extended task has executed.

    A common use-case for extensions is to produce additional artifacts
    from the output of another task. Also, for long-running tasks, it is
    sometimes beneficial to utilize the intermediate output from an extended
    task. The extension artifact can then be acquired more cheaply than if the
    extension had performed all of the work from scratch.

    An extension can only extend one other task.
    """

    fast = False
    """
    Indication of task speed.

    The information is used by the distributed execution strategy to
    optimize how tasks are scheduled. Scheduling tasks remotely is always
    associated with some overhead and sometimes it's beneficial to instead
    schedule fast tasks locally if possible.

    An extended task is only considered fast if all extensions are fast.
    """

    influence = []
    """ List of influence provider objects """

    joltdir = "."
    """ Path to the directory of the .jolt file where the task was defined. """

    joltproject = None
    """ Name of project this task belongs to. """

    name = None
    """ Name of the task. Derived from class name if not set. """

    requires = []
    """ List of dependencies to other tasks. """

    selfsustained = False
    """ Consume this task independently of its requirements.

    Requirements of a self-sustained task will be pruned if the task artifact
    is present in a cache. In other words, if the task is not executed its
    requirements are considered unnecessary.

    For example, consider the task graph A -> B -> C. If B is self-sustained
    and present in a cache, C will never be executed. C will also never be a
    transitive requirement of A. If A requires C, it should be listed
    as an explicit requirement.

    Using this attribute speeds up execution and reduces network
    traffic by allowing the task graph to be reduced.
    """

    taint = None
    """ An arbitrary value used to change the identity of the task.

    It may be hard to remove bad artifacts in a distributed build
    environment. A better method is to taint the task and let the
    artifact be recreated with a different identity.
    """

    weight = 0
    """
    Indication of task execution time.

    The weight is used to optimize the order in which tasks are
    executed using a heuristic scheduling algorithm where ready
    tasks along the critical path are favored.
    """

    _instance = Export(lambda t: str(uuid.uuid4()))
    """
    Task instance identifier.

    The instance ID identifies an execution request and is always unique.

    While the task hash identity may be identical for multiple parallel
    execution requests, the instance ID won't be.
    """

    def __init__(self, parameters=None, manifest=None, **kwargs):
        self._identity = None
        self._report = _JoltTask()
        self.name = self.__class__.name

        self._create_exports_and_parameters()
        self._set_parameters(parameters)

        self.cacheable = self.__class__.cacheable
        self.extends = self.expand(utils.call_or_return_list(self, self.__class__.extends))
        raise_task_error_if(
            len(self.extends) != 1, self,
            "Multiple tasks extended, only one allowed")
        self.extends = self.extends[0]
        self.influence = utils.call_or_return_list(self, self.__class__._influence)
        self.influence.append(TaskClassSourceInfluence())
        self.influence.append(TaintInfluenceProvider())
        self.requires = utils.unique_list(utils.call_or_return_list(self, self.__class__._requires))
        self.selfsustained = utils.call_or_return(self, self.__class__._selfsustained)
        self.tools = Tools(self, self.joltdir)
        self._apply_manifest(manifest)
        self.requires = self.expand(self.requires)

    def _apply_manifest(self, manifest):
        if manifest is None:
            return
        manifest_task = manifest.find_task(self.qualified_name)
        if manifest_task is not None:
            if manifest_task.identity:
                self.identity = manifest_task.identity
            for attrib in manifest_task.attributes:
                export = utils.getattr_safe(self, attrib.name)
                assert isinstance(export, Export), \
                    "'{0}' is not an exportable attribute of task '{1}'"\
                    .format(attrib.name, self.qualified_name)
                export.assign(attrib.value)

    def _influence(self):
        return utils.as_list(self.__class__.influence)

    def _requires(self):
        return utils.call_or_return_list(self, self.__class__.requires)

    def _selfsustained(self):
        return utils.call_or_return(self, self.__class__.selfsustained)

    def _create_exports_and_parameters(self):
        self._exports = {}
        self._parameters = {}

        for key, export in Export.__get_exports__(self).items():
            export = copy.copy(export)
            export.set_task(self)
            setattr(self, key, export)
            self._exports[key] = export

        for key, param in Parameter.__get_params__(self).items():
            param = copy.copy(param)
            setattr(self, key, param)
            self._parameters[key] = param

    def _set_parameters(self, params):
        params = params or {}
        for key, value in params.items():
            try:
                param = utils.getattr_safe(self, key)
            except AttributeError:
                raise_task_error(self, "No such parameter '{0}'", key)
            if isinstance(param, Parameter):
                try:
                    param.set_value(value)
                except ValueError as e:
                    raise_task_error(
                        self,
                        "Illegal value '{0}' assigned to parameter '{1}'",
                        str(e), key)
                continue
            raise_task_error(self, "No such parameter '{0}'", key)
        self._assert_required_parameters_assigned()

    @staticmethod
    def _set_default_parameters(cls, params):
        params = params or {}
        for key, value in params.items():
            param = utils.getattr_safe(cls, key, None)
            if isinstance(param, Parameter):
                param = copy.copy(param)
                param.set_default(value)
                setattr(cls, key, param)
                continue
            raise_task_error(cls.name, "No such parameter '{0}'", key)

    def _assert_required_parameters_assigned(self):
        for key, param in self._get_parameter_objects().items():
            raise_task_error_if(
                param.is_required() and param.is_unset(), self,
                "Required parameter '{0}' has not been set", key)

    def _verify_influence(self, deps, artifact, tools, sources=None):
        # Verify that any transformed sources are influencing
        sources = set(map(tools.expand_path, sources or []))

        # Verify that published files are influencing
        for src, _ in artifact.files.items():
            src = tools.expand_path(src)
            if fs.path.isdir(src):
                sources.update(fs.scandir(src))
            else:
                sources.add(src)

        def _subpath_filter(rootpath):
            def _filter(fname):
                return not fs.is_relative_to(fname, rootpath)
            return _filter

        for _, dep in deps.items():
            deptask = dep.get_task()
            if isinstance(deptask, FileInfluence):
                # Resource dependencies may cover the influence implicitly
                deppath = self.tools.expand_path(str(deptask.path))
                sources = set(filter(lambda d: not deptask.is_influenced_by(self, d), sources))
            else:
                # Ignore any files in artifacts
                deppath = self.tools.expand_path(dep.path)
                sources = set(filter(_subpath_filter(deppath), sources))

        # Ignore any files in build directories
        sources = filter(_subpath_filter(tools.expand_path(tools.buildroot)), sources)
        sources = set(sources)

        for ip in self.influence:
            if not isinstance(ip, FileInfluence):
                continue
            ok = [source for source in sources if ip.is_influenced_by(self, source)]
            sources.difference_update(ok)
        for source in sources:
            log.warning("Missing influence: {} ({})", source, self.name)
        raise_task_error_if(sources, self, "Task is missing source influence")

    def _get_export_objects(self):
        return self._exports

    def _get_parameter_objects(self, unset=False):
        return self._parameters

    def _get_parameters(self, unset=False):
        return {
            key: param.get_value()
            for key, param in self._get_parameter_objects().items()
            if unset or not param.is_unset()
        }

    def _get_explicitly_set_parameters(self):
        return {
            key: param.get_value()
            for key, param in self._get_parameter_objects().items()
            if param.is_set()
        }

    def __str__(self):
        return str(self.name)

    @property
    def canonical_name(self):
        return utils.canonical(self.name)

    @property
    def qualified_name(self):
        return utils.format_task_name(
            self.name,
            self._get_parameters())

    @property
    def short_qualified_name(self):
        return utils.format_task_name(
            self.name,
            self._get_explicitly_set_parameters())

    def expand(self, string_or_list, *args, **kwargs):
        """ Expands keyword arguments/macros in a format string.

        See :func:`jolt.Tools.expand` for details.
        """

        try:
            kwargs["_instance"] = self
            if type(string_or_list) == list:
                return [utils.expand(string, *args, **kwargs) for string in string_or_list]
            return utils.expand(string_or_list, *args, **kwargs)
        except KeyError as e:
            raise_task_error(self, "Invalid macro '{0}' encountered - forgot to set a parameter?", e)

    @property
    def identity(self):
        return self._identity

    @identity.setter
    def identity(self, identity):
        self._identity = identity

    def is_cacheable(self):
        return self.cacheable

    def is_runnable(self):
        return True


class Task(TaskBase):
    #: Path to the .jolt file where the task was defined.

    abstract = True
    """ An abstract task class indended to be subclassed.

    Abstract tasks can't be executed and won't be listed.
    """

    def __init__(self, parameters=None, **kwargs):
        super(Task, self).__init__(parameters, **kwargs)

    def info(self, fmt, *args, **kwargs):
        """
        Log information about the task.
        """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.info(fmt, *args, **kwargs)

    def warning(self, fmt, *args, **kwargs):
        """ Log a warning concerning the task """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.warning(fmt, *args, **kwargs)

    def error(self, fmt, *args, **kwargs):
        """ Log an error concerning the task """
        fmt = self.tools.expand(fmt, *args, **kwargs)
        log.error(fmt, *args, **kwargs)

    def clean(self, tools):
        """
        Cleans up resources and intermediate files created by the task.

        The method is invoked in response to the user running clean
        on the command line. It should restore the environment to its
        original state. The next execution of the task should behave
        as if the task is executed for the first time.

        An implementation must not clean any local or remote artifact cache.
        """

    def run(self, deps, tools):
        """
        Performs the work of the task.

        Dependencies specified with "requires" are passed as the
        deps dictionary. The tools argument provides a set of low
        level tool functions that may be useful.

        .. code-block:: python

          with tools.cwd("path/to/subdir"):
              tools.run("make {target}")

        When using methods from the toolbox, task parameters, such
        as ``target`` above,  are automatically expanded to their values.
        """

    def publish(self, artifact, tools):
        """
        Publishes files produced by :func:`~run`.

        Files can be collected in to the artifact by calling
        artifact.collect().

        Additional metadata can be provided, such as environment
        variables that should be set whenever the task artifact is
        consumed. Example:

        .. code-block:: python

          # Append <artifact-path>/bin to the PATH
          artifact.environ.PATH.append("bin")

          # Pass an arbitrary string to a consumer
          artifact.strings.foo = "bar"

        """

    def unpack(self, artifact, tools):
        """
        Unpacks files published by publish() .

        The intention of this hook is to make necessary adjustments
        to artifact files and directories once they have been downloaded
        into the local cache on a different machine. For example,
        paths may have to be adjusted or an installer executed.

        This hook is executed in the context of a consuming task.
        """
        raise NotImplementedError()

    def shell(self, deps, tools):
        """
        Invoked to start a debug shell.

        The method prepares the environment with attributes exported by task requirement
        artifacts. The shell is entered by passing the ``-g`` flag to the build command.

        Task execution resumes normally when exiting the shell.
        """
        with tools.environ(PS1="jolt$ ") as env:
            from jolt import config
            subprocess.call(config.get_shell().split(), env=env, cwd=tools._cwd)

    @contextmanager
    def report(self):
        """
        Provide error analysis for task.

        Intentionally undocumented. Use at own risk.
        """
        yield ReportProxy(self, self._report)


class SubTask(object):
    def __init__(self, task):
        self._deps = []
        self._identity = []
        self._influence = []
        self._message = None
        self._outputs = []
        self._task = task
        self._tools = Tools(task, task.joltdir)

    def __str__(self):
        if self.message:
            return self.message
        if self._outputs:
            return " ".join(self._outputs)
        return None

    @property
    def dependencies(self):
        return self._deps

    @functools.cached_property
    def identity(self):
        sha = hashlib.sha1()
        for ident in self._identity:
            sha.update(ident.encode())
        for output in self._outputs:
            sha.update(output.encode())
        if self.message:
            sha.update(self.message.encode())
        return sha.hexdigest()

    @functools.cached_property
    def influence(self):
        sha = hashlib.sha1()
        for infl in self._influence:
            if callable(infl):
                sha.update(infl().encode())
            else:
                sha.update(infl.encode())
        for dep in self._deps:
            sha.update(dep.influence.encode())
        return sha.hexdigest()

    @functools.cached_property
    def is_outdated(self):
        try:
            with self._tools.cwd(self._tools.builddir("subtasks", incremental=True)):
                if self.influence != self._tools.read_file(self.identity):
                    return True
                for output in self.outputs:
                    output = fs.as_canonpath(output)
                    assert self._tools.read_file(output + ".identity") == self.identity

                # FIXME: Check hash content of output file
                for output in self.outputs:
                    if not fs.path.exists(fs.path.join(self._task.joltdir, output)):
                        return True
                for dep in self.dependencies:
                    if dep.is_outdated:
                        return True
            return False
        except Exception:
            return True

    def set_uptodate(self):
        try:
            del self.influence
            del self.is_outdated
        except AttributeError:
            pass
        with self._tools.cwd(self._tools.builddir("subtasks", incremental=True)):
            self._tools.write_file(self.identity, self.influence)

            for output in self.outputs:
                output = fs.as_canonpath(output)
                self._tools.mkdirname(output)
                self._tools.write_file(output + ".identity", self.identity)

    def run(self):
        pass

    def add_dependency(self, dep):
        deps = utils.as_list(dep)
        subtasks = [self._task._add_input(dep) for dep in deps]
        self._deps.extend(subtasks)

    def add_identity(self, infl):
        self._identity.append(infl)

    def add_influence(self, infl):
        self._influence.append(infl)

    def add_influence_file(self, path):
        self.add_influence(utils.filesha1(path))

    def add_influence_depfile(self, path):
        def depfile():
            result = ""
            deps = self._tools.read_depfile(fs.path.join(self._task.joltdir, path))
            for output in self.outputs:
                for input in deps.get(output, []):
                    result += utils.filesha1(input)
            return result
        self.add_influence(depfile)

    def add_output(self, output):
        self.add_identity(output)
        if type(output) == list:
            self._outputs.extend(output)
        else:
            self._outputs.append(output)

    @property
    def message(self):
        return self._message

    def set_message(self, message):
        self._message = message

    @property
    def outputs(self):
        return self._outputs


class Input(SubTask):
    def __init__(self, task, input):
        super().__init__(task)
        self._outputs = [input]
        self.add_influence_file(input)
        self.set_uptodate()

    @property
    def message(self):
        return f"[IN] {self._outputs[0]}"

    def run(self):
        raise_task_error(self._task, "Input file '{}' does not exist", self._outputs[0])


class CommandSubtask(SubTask):
    def __init__(self, task, command):
        super().__init__(task)
        self.add_identity(command)
        self._command = command

    def __str__(self):
        s = super().__str__()
        return s if s is not None and not log.is_verbose() else self._command

    def run(self):
        self._tools.run(self._command)


class FunctionSubtask(SubTask):
    def __init__(self, task, fn):
        super().__init__(task)
        self.add_identity(fn.__name__)
        self.fn = fn

    def run(self):
        self.fn(self)


class RenderSubtask(SubTask):
    def __init__(self, task, template, **kwargs):
        super().__init__(task)
        self._data = self._tools.render(template, **kwargs)
        self.add_identity(template)
        self.add_influence(self._data)

    def run(self):
        for output in self.outputs:
            self._tools.write_file(output, self._data, expand=False)


class FileRenderSubtask(SubTask):
    def __init__(self, task, path, **kwargs):
        super().__init__(task)
        self.add_identity(path)
        self._path = path

    def run(self):
        data = self._tools.render_file(self._path)
        for output in self.outputs:
            self._tools.write_file(output, data, expand=False)


class MultiTask(Task):
    """
    A task with subtasks that are executed in parallel with intermediate caching.

    A MultiTask is useful for tasks with many subtasks that benefit from intermediate
    caching, such as compilation tasks where multiple source files are compiled into
    object files and then either linked into an executable or archived into a library.

    Subtasks are executed in parallel and their output is cached locally in a build
    directory. The output is not automatically shared with other Jolt clients, only
    the files published by the MultiTask is shared. A subtask is only re-executed
    if the influence one of its dependencies change.

    Subtasks are defined in the MultiTask method generate() and they can be either
    a shell command or a python function. Helper methods in the class allow
    implementors to define outputs and inter-subtask dependencies.

    Example:

        .. code-block:: python

          flags = ["-DDEBUG"]

          def generate(self, deps, tools):
              sources = ["a.cpp", "b.cpp", "c.cpp"]
              objects = []

              # Create compilation subtasks for each source file
              for source in sources:
                  object = self.command(
                      "g++ {flags} -c {inputs} -o {outputs} ",
                      inputs=[source],
                      outputs=[source  +".o"])
                  objects.append(object)

              # Create linker subtask
              executable = self.command(
                  "g++ {inputs} -o {output}",
                  inputs=objects,
                  outputs=["executable"])

    """

    abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subtasks = []
        self._subtasks_by_output = {}

    def _add_subtask(self, subtask):
        self._subtasks.append(subtask)

        for output in subtask.outputs:
            if output in self._subtasks_by_output:
                othersubtask = self._subtasks_by_output
                raise_task_error(
                    self,
                    "'{}' is generated by both '{}' and '{}'",
                    output, subtask.command, othersubtask.command)
            self._subtasks_by_output[output] = subtask

        return subtask

    def _find_subtask(self, output):
        return self._subtasks_by_output.get(output)

    def _add_input(self, input):
        raise_task_error_if(not input, self, "Input is None")
        if not isinstance(input, SubTask):
            inputsubtask = self._find_subtask(input)
            if inputsubtask:
                return inputsubtask
            inputsubtask = Input(self, input)
        else:
            inputsubtask = input
        self._subtasks.append(inputsubtask)
        self._subtasks_by_output[input] = inputsubtask
        return inputsubtask

    def _to_subtask_list(self, inputs):
        inputs = utils.as_list(inputs)
        return [self._add_input(input) for input in inputs]

    def _to_output_files(self, subtasks):
        subtasks = utils.as_list(subtasks)
        outputs = []
        for subtask in subtasks:
            outputs.extend(subtask.outputs)
        return outputs

    def _to_input_subtasks(self, inputs, **kwargs):
        inputs = utils.as_list(inputs)
        subtasks = []
        for input in inputs:
            input = self.expand(input, **kwargs)
            subtasks.append(self._add_input(input))
        return subtasks

    def command(self, command, inputs=None, outputs=None, message=None, mkdir=True, **kwargs):
        """
        Create shell command subtask.

        The subtask executes the specified command. String format specifiers may be used and
        are resolved primarily by kwargs and secondarily by task attributes.

        Args:
            inputs (str, list): files or subtasks that the command depends on.
            outputs (str, list): list of files that the subtasks produces.
            message (str): custom message that the subtask will print when executed.
            mkdir (boolean): automatically create directories for outputs. If False,
                the caller must ensure that the directories exist before the the subtask
                is executed.
            kwargs: additional keyword values used to format the command line string.

        Returns:
            Subtask object.

        Example:

            .. code-block:: python

              executable = self.command(
                  "g++ {inputs} -o {output}",
                  inputs=["main.cpp"],
                  outputs=["executable"])

        """
        inputs = self._to_input_subtasks(inputs, **kwargs)
        inputfiles = self._to_output_files(inputs)

        outputs = utils.as_list(outputs)
        outputs = [self.expand(output, **kwargs) for output in outputs]
        outputs = [fs.path.relpath(output) for output in outputs]

        dirs = set()
        if mkdir:
            for output in outputs:
                dirs.add(self.mkdirname(output, inputs=inputfiles, outputs=outputs, **kwargs))

        command = self.expand(command, inputs=inputfiles, outputs=outputs, **kwargs)
        subtask = CommandSubtask(self, command)

        for dir in dirs:
            subtask.add_dependency(dir)
        for input in inputs:
            subtask.add_dependency(input)
        for output in outputs:
            output = self.expand(output, inputs=inputfiles, outputs=outputs, **kwargs)
            subtask.add_output(output)

        if message:
            subtask.set_message(self.expand(message, inputs=inputfiles, outputs=outputs, **kwargs))

        self._add_subtask(subtask)

        return subtask

    def call(self, fn, outputs, **kwargs):
        """
        Create a Python function call subtask.

        The subtask executes the specified Python function, passing the subtask as argument.

        Args:
            fn (func): Python function to execute.
            outputs (str, list): list of files that the subtasks produces.
            kwargs: additional keyword values used to format the output file paths.

        Returns:
            Subtask object.

        Example:

            .. code-block:: python

              def mkdir(subtask):
                  for output in subtask.outputs:
                      self.tools.mkdir(output)

              dirtask = self.call(mkdir, outputs=["newly/created/directory"])

        """
        subtask = FunctionSubtask(self, fn)
        outputs = utils.as_list(outputs)
        for output in outputs:
            output = self.tools.expand_relpath(output, self.joltdir, outputs=outputs, **kwargs)
            subtask.add_output(output)
        self._add_subtask(subtask)
        return subtask

    def mkdir(self, path, *args, **kwargs):
        """
        Create a subtask that creates a directory.

        Args:
            path (str): Path to directory.
            kwargs: additional keyword values used to format the directory path string.

        Returns:
            Subtask object.

        Example:

            .. code-block:: python

              dirtask = self.mkdir("{outdir}/directory", outdir=tools.builddir())
        """
        path = self.expand(path, *args, **kwargs)
        subtask = self._find_subtask(path)
        if not subtask:
            subtask = self.call(lambda subtask: self.tools.mkdir(path), [path])
        return subtask

    def mkdirname(self, path, *args, **kwargs):
        """
        Create a subtask that creates a parent directory.

        Args:
            path (str): Path for which the parent directory shall be created.
            kwargs: additional keyword values used to format the directory path string.

        Returns:
            Subtask object.

        Example:

            .. code-block:: python

              # Creates {outdir}/directory
              dirtask = self.mkdir("{outdir}/directory/object.o", outdir=tools.builddir())

        """
        path = self.expand(path, *args, **kwargs)
        path = fs.path.dirname(path)
        return self.mkdir(path, *args, **kwargs)

    def render(self, template, outputs, **kwargs):
        """
        Create a subtask that renders a Jinja template string to file.

        Args:
            template (str): Jinja template string.
            outputs (str, list): list of files that the subtasks produces.
            kwargs: additional keyword values used to render the template and output file paths.

        Returns:
            Subtask object.

        Example:

            .. code-block:: python

              # Creates file.list with two lines containing "a.o" and "b.o"

              template_task = self.render(
                  "{% for line in lines %}{{ line }}\\n{% endfor %}",
                  outputs=["file.list"],
                  lines=["a.o", "b.o"])

        """
        outputs = utils.as_list(outputs)
        subtask = RenderSubtask(self, template, **kwargs)
        for output in outputs:
            output = self.tools.expand_relpath(output, self.joltdir, outputs=outputs, **kwargs)
            subtask.add_output(output)
        self._add_subtask(subtask)
        return subtask

    def render_file(self, template, outputs, **kwargs):
        """
        Create a subtask that renders a Jinja template file to file.

        Args:
            template (str): Jinja template file path.
            outputs (str, list): list of files that the subtasks produces.
            kwargs: additional keyword values used to format the output file paths.

        Returns:
            Subtask object.

        Example:

            .. code-block:: python

              # Render file.list.template into file.list
              template_task = self.render_file("file.list.template", outputs=["file.list"])

        """
        template = self._to_subtask_list(template)
        templatefiles = self._to_output_files(template)
        raise_task_error_if(len(templatefiles) > 1, "Can only render one template at a time")

        outputs = utils.as_list(outputs)
        subtask = FileRenderSubtask(self, templatefiles[0], **kwargs)
        for output in outputs:
            output = self.tools.expand_relpath(output, outputs=outputs, **kwargs)
            subtask.add_output(output)
        for input in templatefiles:
            subtask.add_dependency(input)
        self._add_subtask(subtask)
        return subtask

    def generate(self, deps, tools):
        """
        Called to generate subtasks.

        An implementer can override this method in order to create subtasks
        that will later be executed during the :func:`~run` stage of the task.

        Subtasks can be defined using either of these helper methods:

          - :func:`~call`
          - :func:`~command`
          - :func:`~mkdir`
          - :func:`~mkdirname`
          - :func:`~render`
          - :func:`~render_file`

        """
        pass

    def run(self, deps, tools):
        """
        Executes subtasks defined in :func:`~generate`.

        This method should typically not be overridden in subclasses.
        """

        self.generate(deps, tools)

        log.debug("About to start executing these subtasks:")
        for subtask in self._subtasks:
            for subtaskout in subtask.outputs:
                if not subtask.dependencies:
                    log.debug("  {}", subtaskout)
                for dep in subtask.dependencies:
                    for depout in dep.outputs:
                        log.debug("  {}: {}", subtaskout, depout)

        subtasks = {}
        deps = {}

        # Build graph of inverse dependencies
        for subtask in self._subtasks:
            if subtask not in subtasks:
                subtasks[subtask] = []
                deps[subtask] = []
            for dep in subtask.dependencies:
                if dep not in deps:
                    deps[dep] = []
                deps[dep].append(subtask)
                subtasks[subtask].append(dep)

        # Prune up-to-date subtasks
        for subtask in list(filter(lambda subtask: not subtask.is_outdated, subtasks.keys())):
            del subtasks[subtask]
            for dep in deps[subtask]:
                subtasks[dep].remove(subtask)

        subtaskindex = 0
        subtaskcount = len(subtasks)

        with ThreadPoolExecutor(max_workers=tools.cpu_count()) as pool:
            futures = {}

            while subtasks or futures:
                completed = []
                candidates = [subtask for subtask, deps in subtasks.items() if not deps]

                if not candidates and not futures:
                    break

                for subtask in candidates:
                    subtaskindex += 1
                    del subtasks[subtask]
                    if subtask.is_outdated:
                        log.info("[{}/{}] {}", subtaskindex, subtaskcount, str(subtask))
                        futures[pool.submit(subtask.run)] = subtask
                    else:
                        completed.append(subtask)

                for future in as_completed(futures.keys()):
                    subtask = futures[future]
                    del futures[future]
                    completed.append(subtask)

                    try:
                        future.result()
                        subtask.set_uptodate()
                    except Exception as e:
                        for future in futures:
                            future.cancel()
                        raise e
                    break

                for subtask in completed:
                    for dep in deps[subtask]:
                        subtasks[dep].remove(subtask)

            if subtasks:
                log.debug("These remaining subtasks could not be started due to unresolved dependencies")
                for subtask in subtasks:
                    log.debug("  {}", str(subtask))
                    for dep in subtask.dependencies:
                        log.debug("   - {}", str(dep))

            raise_task_error_if(subtasks, self, "Subtasks with unresolved dependencies could not be executed")


class ErrorProxy(object):
    def __init__(self, error):
        self._error = error


class ReportProxy(object):
    def __init__(self, task, report):
        self._task = task
        self._report = report

    def add_error(self, type, location, message, details=""):
        """ Add an error to the build report. """
        error = self._report.create_error()
        error.type = type
        error.location = location
        error.message = message
        error.details = details
        return error

    def add_regex_errors(self, type, regex, logbuf):
        """
        Find errors in log using regex and add them to build report.

        The regex should contain these named match groups:

         - location - origin of the error
         - message  - a brief description of the error
         - details  - futher error details

        """
        for match in re.finditer(regex, logbuf):
            error = match.groupdict()
            self.add_error(
                type,
                error.get("location", ""),
                error.get("message", ""),
                error.get("details", ""))

    def add_regex_errors_with_file(self, type, regex, logbuf, reldir, filterfn=lambda n: True):
        """
        Find errors in log using regex and add them to build report.

        Instead of using error details from the regex match,
        file content is used. For this to work the regex match
        must contain these named groups:

         - file - path to file
         - line - line number of error

        In case file is a relative path, reldir is the working directory.
        """
        errors_by_location = OrderedDict()
        for match in re.finditer(regex, logbuf):
            error = match.groupdict()
            if not filterfn(error):
                continue
            if error["location"] not in errors_by_location:
                errors_by_location[error["location"]] = (error, [error["message"]])
            else:
                errors_by_location[error["location"]][1].append(error["message"])

        for error, msgs in errors_by_location.values():
            message = "\n".join(utils.unique_list(msgs))
            with self._task.tools.cwd(reldir):
                try:
                    details = self._task.tools.read_file(error["file"])
                    details = details.splitlines()
                    details = str(error["line"]) + ": " + details[int(error["line"]) - 1]
                except Exception:
                    details = ""
            self.add_error(type, error.get("location", ""), message, details)

    def add_exception(self, exc, errtype=None, location=None):
        """
        Add an exception to the build report.

        The exception traceback is included in the error details, but
        frames inside Jolt may be filtered out if the exception
        originated in a task recipe.

        No traceback is included if the exception is derived from JoltError.

        """
        tb = traceback.format_exception(type(exc), value=exc, tb=exc.__traceback__)
        installdir = fs.path.dirname(__file__)
        if any(map(lambda frame: installdir not in frame, tb[1:-1])):
            while len(tb) > 2 and installdir in tb[1]:
                del tb[1]
        loc = re.findall("\"(.*?\", line [0-9]+, in .*?)\n", tb[1])
        location = location or (loc[0] if loc and len(loc) > 0 else "")
        message = str(exc)
        if isinstance(exc, JoltCommandError):
            details = "\n".join(exc.stderr)
        elif isinstance(exc, JoltError):
            details = ""
        else:
            details = "".join(tb)

        self.add_error(
            type=errtype or ("Exception" if not isinstance(exc, JoltError) else "Error"),
            location=location,
            message=message,
            details=details)

    @property
    def errors(self):
        return [ErrorProxy(error) for error in self._report.errors]

    @property
    def manifest(self):
        return self._report


class Resource(Task):
    """
    A resource task.

    Resources are special tasks executed in the :class:`~jolt.Context` of other tasks.
    They are invoked to acquire and release a resource, such as hardware equipment,
    before and after the execution of a task. No artifact is produced by a resource.

    Implementors should override :func:`~acquire` and :func:`~release`.

    """

    cacheable = False

    abstract = True
    """ An abstract resource class indended to be subclassed. """

    def __init__(self, *args, **kwargs):
        super(Resource, self).__init__(*args, **kwargs)

    def is_runnable(self):
        return False

    def info(self, fmt, *args, **kwargs):
        pass

    def acquire(self, artifact, deps, tools, owner):
        """ Called to acquire the resource.

        An implementor overrides this method in a subclass. The acquired
        resource must be released manually if an exception occurs before the
        method has returned. If this method returns successfully, the :func:`~release`
        method is guaranteed to be called in the future upon completion of the
        consuming task (unless the process is forcibly interrupted or killed).

        Args:
            artifact (:class:`~jolt.Artifact`): The artifact associated with the resource.
                It is not possible to publish files from a resource, but the implementor
                can still use the resource to pass information to consuming tasks.
            deps (:class:`~jolt.Context`): Task execution context used to access the
                artifacts of dependencies.
            tools (:class:`~jolt.Tools`): A task specific toolbox.
            owner (:class:`~jolt.Task`): The owner task for which the resource is acquired.

        """

    def release(self, artifact, deps, tools, owner):
        """ Called to release the resource.

        An implementor overrides this method in a subclass.

        Args:
            artifact (:class:`~jolt.Artifact`): The artifact associated with the resource.
                It is not possible to publish files from a resource, but the implementor
                can still use the resource to pass information to consuming tasks.
            deps (:class:`~jolt.Context`): Task execution context used to access the
                artifacts of dependencies.
            tools (:class:`~jolt.Tools`): A task specific toolbox.
            owner (:class:`~jolt.Task`): The owner task for which the resource is released.
        """

    def run(self, env, tools):
        self._run_env = env


class WorkspaceResource(Resource):
    """
    A workspace resource task.

    A workspace resource is a similiar to a regular resource and may be used as such,
    but it can also be used from a workspace manifest to acquire project resources
    before any task recipes are parsed. A common use-case is to retrieve project
    sources from SCM. Workspace resources cannot have any dependencies.
    No artifact is produced.

    Implementors should override :func:`~acquire_ws` and :func:`~release_ws`.
    """

    def __init__(self, *args, **kwargs):
        super(WorkspaceResource, self).__init__(*args, **kwargs)
        raise_task_error_if(len(self.requires) > 0, self,
                            "Workspace resource is not allowed to have requirements")

    def acquire(self, **kwargs):
        return self.acquire_ws()

    def release(self, **kwargs):
        return self.release_ws()

    def acquire_ws(self):
        """ Called to acquire the resource.

        An implementor overrides this method in a subclass. The acquired
        resource must be released manually if an exception occurs before the
        method has returned. """

    def release_ws(self):
        """ Called to release the resource.

        An implementor overrides this method in a subclass.

        """


@attributes.requires("_image")
class Chroot(Resource):
    """
    Resource to use task artifact or directory path as chroot in consumers.

    Example:

      .. code-block:: python

        from jolt import Chroot, Task
        from jolt.plugins.podman import ContainerImage

        class SdkImage(ContainerImage):
            dockerfile = \"\"\"
            FROM debian:sid-slim
            ARG DEBIAN_FRONTEND=noninteractive
            RUN apt-get update && apt-get install -y --no-install-recommends gcc g++ && apt-get clean
            \"\"\"
            output = "directory"

        class Sdk(Chroot):
            chroot = "sdkimage"

        class Compile(Task):
            requires = ["sdk"]

            def run(self, deps, tools):
                tools.run("gcc -v")

    """

    chroot = None
    """ Task name or directory path to use as chroot """

    @property
    def _image(self):
        registry = TaskRegistry.get()
        if registry.get_task_class(self.expand(self.chroot)):
            return [self.chroot]
        return []

    def acquire(self, artifact, deps, tools, owner):
        try:
            rootfs = deps[self.chroot]
        except Exception:
            rootfs = tools.expand(self.image)
        self._context_stack = ExitStack()
        self._context_stack.enter_context(
            owner.tools.chroot(rootfs))

    def release(self, artifact, deps, tools, owner):
        self._context_stack.close()


class Alias(Task):
    """
    An alias task.

    Aliases are a special kind of task which can be used to introduce new
    names for other tasks or groups of tasks. They are useful as milestones
    when building continuous integration pipelines since they won't
    be executed, thus saving time compared to a regular task.
    """

    cacheable = False

    abstract = True
    """ An abstract alias class indended to be subclassed. """

    def __init__(self, *args, **kwargs):
        super(Alias, self).__init__(*args, **kwargs)
        raise_task_error_if(
            self.extends, self, "Aliases cannot be extensions")

    def is_runnable(self):
        return False

    def info(self, fmt, *args, **kwargs):
        pass


class Download(Task):
    """
    Downloads file(s) over HTTP(S).

    Once downloaded, archives are extracted and all of their files are published.
    If the file is not an archive it is published as is. Recognized archive extensions are:

        - .tar
        - .tar.bz2
        - .tar.gz
        - .tar.xz
        - .tgz
        - .zip

    Example:

      .. code-block:: python

        class NodeJS(Download):
            \"\"\" Downloads and publishes Node.js. Adds binaries to PATH. \"\"\"

            version = Parameter("14.16.1")
            url = "https://nodejs.org/dist/v{version}/node-v{version}-win-x64.zip"

            def publish(self, artifact, tools):
                super(publish).publish(artifact, tools)
                artifact.environ.PATH.append("node-v{version}-win-x64")
    """

    abstract = True

    collect = ["*"]
    """
    A list of file publication instructions.

    Items in the list are passed directly to :func:`Artifact.collect() <jolt.Artifact.collect>`
    and can be either strings, tuples or dictionaries.

    Example:

      .. code-block:: python

            collect = [
                "*",                           # Collect all files
                ("*", "src/"),                 # Collect all files into the artifact's src/ directory
                {"files": "*", cwd="subdir"},  # Collect all files from the archive's subdir/ directory
            ]

    """

    extract = True
    """ Automatically extract archives. """

    url = None
    """
    URL(s) of file(s) to download.

    A single URL string is accepted, as well as a list of URL strings.
    """

    symlinks = True
    """ Publish symlinks (True) """

    def _filename_from_url(self, tools, url):
        from urllib.parse import urlparse
        url = urlparse(tools.expand(url))
        return fs.posixpath.basename(url.path) or "file"

    def run(self, deps, tools):
        supported_formats = [".tar", ".tar.bz2", ".tar.gz", ".tar.xz", ".tgz", ".zip"]

        raise_task_error_if(not self.url, self, "No URL(s) specified")

        self._downloaddir = tools.builddir()
        self._extractdir = tools.builddir("extracted")
        for url in utils.as_list(self.url):
            filename = self._filename_from_url(tools, url)
            with tools.cwd(self._downloaddir):
                tools.download(url, filename)

            if self.extract and any(map(lambda n: filename.endswith(n), supported_formats)):
                with tools.cwd(self._extractdir):
                    self.info("Extracting {}", filename)
                    tools.extract(fs.path.join(self._downloaddir, filename), ".")
            else:
                with tools.cwd(self._downloaddir):
                    tools.copy(filename, self._extractdir)

    def publish(self, artifact, tools):
        with tools.cwd(self._extractdir):
            for files in self.collect:
                if type(files) == tuple:
                    artifact.collect(*files, symlinks=self.symlinks)
                elif type(files) == dict:
                    artifact.collect(**files, symlinks=self.symlinks)
                else:
                    artifact.collect(files, symlinks=self.symlinks)


class Script(Task):
    """
    A simple shell script task.

    The script source is extracted directly from the task class documentation.
    All text following a ``---`` separator will be executed.

    A temporary build directory is automatically created and can be accessed
    with ``{builddir}``. All other task attributes are also expanded as usual.
    Dependency artifacts are accessible through the ``deps`` dictionary.

      .. code-block:: python

        echo {deps[task].path}

    By default, all files in the build directory are published in the task artifact.
    The :attr:`collect` attribute can be used to customize file collection.
    Alternatively, the publish method may be overridden.

    Keep in mind that shell scripts are not necessarily portable between
    host operating systems. Implement your tasks in Python code if portability
    is a concern.

    Example:

      .. code-block:: python

        class Hello(Script):
            \"\"\"
            Classic Hello World!
            ---
            # Script source

            echo "Hello world!" > {builddir}/hello.txt
            \"\"\"
    """
    abstract = True

    collect = [{"files": "*", "cwd": "{builddir}"}]
    """
    A list of file publication instructions.

    Items in the list are passed directly to :func:`Artifact.collect() <jolt.Artifact.collect>`
    and can be either strings, tuples or dictionaries.

    By default, all files in the build directory are published.

    Example:

      .. code-block:: python

        collect = [
            "*",                              # Collect all files
            ("*", "src/"),                    # Collect all files into the artifact's src/ directory
            {"files": "*", "cwd": "subdir"},  # Collect all files from the archive's subdir/ directory
        ]

    """

    source = None

    def _source(self, tools):
        if self.source is not None:
            return self.source
        doc = self.__doc__.split("---", 1)
        script = doc[1] if len(doc) > 1 else doc[0]
        script = script.splitlines()
        script = [line[4:] for line in script]
        script = "\n".join(script)
        script = script.lstrip()
        if not script.startswith("#!"):
            script = "#!" + tools.getenv("SHELL", "/bin/sh") + "\n" + script
        return script

    def run(self, deps, tools):
        self.builddir = tools.builddir()
        self.deps = deps
        self._scriptdir = tools.builddir("script")

        tools.write_file("{_scriptdir}/script", self._source(tools))
        tools.chmod("{_scriptdir}/script", 0o555)
        tools.run("{_scriptdir}/script")

    def publish(self, artifact, tools):
        with tools.cwd(self.builddir):
            for files in self.collect:
                if type(files) == tuple:
                    artifact.collect(*files)
                elif type(files) == dict:
                    artifact.collect(**files)
                else:
                    artifact.collect(files)


__unittest = True


class _TestCase(ut.FunctionTestCase):
    def __init__(self, task, deps, tools, testfunc, *args, **kwargs):
        super().__init__(testfunc, *args, **kwargs)
        self.task = task
        self.deps = deps
        self.tools = tools
        self.testfunc = testfunc

    @property
    def name(self):
        return self.testfunc.__name__

    @property
    def fullname(self):
        return "{}.{}".format(self.task.__class__.__name__, self.name)

    def runTest(self):
        self.task._run(self)

    def __str__(self):
        return "{} ({})".format(self.task.__class__.__name__, self.testfunc.__name__)

    @property
    def __doc__(self):
        return self.testfunc.__doc__


class _TestResult(ut.TextTestResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.successes = []
        self.errors_exc = []

    def _begin(self):
        log.stdout(ut.TextTestResult.separator1)

    def _end(self):
        log.stdout(ut.TextTestResult.separator2)
        log.stdout("")

    def startTest(self, test):
        self._begin()
        super().startTest(test)
        log.stdout(ut.TextTestResult.separator2)

    def addSuccess(self, test):
        super().addSuccess(test)
        self.successes.append(test)
        self._end()

    def addError(self, test, err):
        _, exc, tb = err
        self.errors_exc.append((test, exc))
        super().addError(test, err)
        self._end()

    def addFailure(self, test, err):
        super().addFailure(test, err)
        self._end()

    def addSkip(self, test, reason):
        super().addSkip(test, reason)
        self._end()


class Test(Task):
    """ A test task.

    The test task combines a regular Jolt task with a Python ``unittest.TestCase``.
    As such, a test task is a collection of similar test-cases where each test-case
    is implemented as an instancemethod named with a ``test_`` prefix. When executed,
    the task runs all test-case methods and summarizes the result.

    All regular ``unittest`` assertions and decorators can be used in the test methods.
    For details about inherited task attributes, see :class:`jolt.tasks.Task` and
    Python unittest.TestCase.

    Example:

    .. code-block:: python

      class OperatorTest(Test):

          def test_add(self):
              self.assertEqual(1+1, 2)

          def test_sub(self):
              self.assertEqual(2-1, 1)

    """

    abstract = True
    """ An abstract test class indended to be subclassed.

    Abstract test tasks can't be executed and won't be listed.
    """

    pattern = Parameter(required=False, help="Test-case filter wildcard.")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def parameterized(args):
        """
        Parameterizes a test method.

        The test method is instantiated and called once with each argument tuple in the list.

        Example:

          .. code-block:: python

            class Example(Test):
               @Test.parameterized([
                   (1, 1, 1),  # 1*1 == 1
                   (1, 2, 2),  # 1*2 == 2
                   (2, 1, 2),  # 2*1 == 2
                   (2, 2, 4),  # 2*2 == 4
               ])
               def test_multiply(self, factor1, factor2, product):
                   self.assertEqual(factor1*factor2, product)

        """
        raise_error_if(type(args) != list, "Test.parameterized() expects a list as argument")

        class partialmethod(functools.partialmethod):
            def __init__(self, index, func, *args):
                super().__init__(func, *args)
                self.__index = index

            def __get__(self, obj, cls=None):
                retval = super().__get__(obj, cls)
                retval.__name__ = f"{self.func.__name__}[{self.__index}]"
                retval.__doc__ = self.func.__doc__
                return retval

        def decorate(method):
            frame = sys._getframe().f_back.f_locals
            for index, arg in enumerate(args):
                testmethod = partialmethod(index, method, *utils.as_list(arg))
                name = f"{method.__name__}[{index}]"
                frame[name] = testmethod
            return None
        return decorate

    def setup(self, deps, tools):
        """ Implement this method to make preparations before a test """

    def cleanup(self):
        """ Implement this method to clean up after a test """

    def skip(self, reason=""):
        raise ut.SkipTest(self.tools.expand(reason))

    def _get_test_names(self):
        return [attrib for attrib in dir(self) if attrib.startswith("test_")]

    def _setup(self, test):
        self._curtest = test
        self._testMethodName = test.testfunc.__name__
        self._testMethodDoc = test.testfunc.__doc__
        self.setup(test.deps, test.tools)

    def _test(self, test):
        test.testfunc()

    def _cleanup(self, test):
        self.cleanup()

    def _run(self, test):
        try:
            self._setup(test)
            self._test(test)
        finally:
            self._cleanup(test)

    def assertTrue(self, *args, **kwargs):
        return self._curtest.assertTrue(*args, **kwargs)

    def assertFalse(self, *args, **kwargs):
        return self._curtest.assertFalse(*args, **kwargs)

    def assertIn(self, *args, **kwargs):
        return self._curtest.assertIn(*args, **kwargs)

    def assertNotIn(self, *args, **kwargs):
        return self._curtest.assertNotIn(*args, **kwargs)

    def assertIs(self, *args, **kwargs):
        return self._curtest.assertIs(*args, **kwargs)

    def assertIsNot(self, *args, **kwargs):
        return self._curtest.assertIsNot(*args, **kwargs)

    def assertIsInstance(self, *args, **kwargs):
        return self._curtest.assertIsInstance(*args, **kwargs)

    def assertIsNotInstance(self, *args, **kwargs):
        return self._curtest.assertIsNotInstance(*args, **kwargs)

    def assertIsNone(self, *args, **kwargs):
        return self._curtest.assertIsNone(*args, **kwargs)

    def assertIsNotNone(self, *args, **kwargs):
        return self._curtest.assertIsNotNone(*args, **kwargs)

    def assertEqual(self, *args, **kwargs):
        return self._curtest.assertEqual(*args, **kwargs)

    def assertNotEqual(self, *args, **kwargs):
        return self._curtest.assertNotEqual(*args, **kwargs)

    def assertAlmostEqual(self, *args, **kwargs):
        return self._curtest.assertAlmostEqual(*args, **kwargs)

    def assertAlmostNotEqual(self, *args, **kwargs):
        return self._curtest.assertAlmostNotEqual(*args, **kwargs)

    def assertGreater(self, *args, **kwargs):
        return self._curtest.assertGreater(*args, **kwargs)

    def assertGreaterEqual(self, *args, **kwargs):
        return self._curtest.assertGreaterEqual(*args, **kwargs)

    def assertLess(self, *args, **kwargs):
        return self._curtest.assertLess(*args, **kwargs)

    def assertLessEqual(self, *args, **kwargs):
        return self._curtest.assertLessEqual(*args, **kwargs)

    def assertRaises(self, *args, **kwargs):
        return self._curtest.assertRaises(*args, **kwargs)

    def assertRaisesRegex(self, *args, **kwargs):
        return self._curtest.assertRaisesRegex(*args, **kwargs)

    def assertRegex(self, *args, **kwargs):
        return self._curtest.assertRegex(*args, **kwargs)

    def assertCountEqual(self, *args, **kwargs):
        return self._curtest.assertCountEqual(*args, **kwargs)

    def skipTest(self, *args, **kwargs):
        return self._curtest.skipTest(*args, **kwargs)

    def subTest(self, *args, **kwargs):
        return self._curtest.subTest(*args, **kwargs)

    def run(self, deps, tools):
        testsuite = ut.TestSuite()
        for test in self._get_test_names():
            if self.pattern.is_unset() or fnmatch.fnmatch(test, str(self.pattern)):
                testfunc = getattr(self, test)
                if not testfunc:
                    continue
                testsuite.addTest(
                    _TestCase(self, deps, tools, testfunc))
        with log.stream() as logstream:
            self.testresult = ut.TextTestRunner(resultclass=_TestResult, stream=logstream, verbosity=2).run(testsuite)
        with self.report() as report:
            for tc, exc in self.testresult.errors_exc:
                report.add_exception(exc, "Test Error", tc.name)
            for tc, tb in self.testresult.failures:
                report.add_error("Test Failed", tc.name, tb.splitlines()[-1], tb)
        raise_unreported_task_error_if(
            not self.testresult.wasSuccessful(), self,
            "{} tests out of {} were successful".format(
                len(self.testresult.successes),
                self.testresult.testsRun))


@ArtifactAttributeSetProvider.Register
class ResourceAttributeSetProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        pass

    def parse(self, artifact, content):
        pass

    def format(self, artifact, content):
        pass

    def apply(self, task, artifact):
        resource = artifact.get_task()
        if isinstance(resource, Resource):
            from inspect import signature

            deps = resource._run_env
            deps.__enter__()
            sig = signature(resource.acquire)
            try:
                ba = sig.bind_partial(artifact=artifact, deps=deps, tools=resource.tools, owner=task)
                acquire = resource.acquire
            except Exception:
                ba = sig.bind_partial(artifact, deps, resource.tools)
                acquire = utils.deprecated(resource.acquire)
            acquire(*ba.args, **ba.kwargs)

    def unapply(self, task, artifact):
        resource = artifact.get_task()
        if isinstance(resource, Resource):
            from inspect import signature

            deps = resource._run_env
            sig = signature(resource.release)
            try:
                ba = sig.bind_partial(artifact=artifact, deps=deps, tools=resource.tools, owner=task)
                release = resource.release
            except Exception:
                ba = sig.bind_partial(artifact, deps, resource.tools)
                release = utils.deprecated(resource.release)
            release(*ba.args, **ba.kwargs)
            deps.__exit__(None, None, None)



import keyring
import getpass
from ftplib import FTP, FTP_TLS
from requests.exceptions import RequestException


from jolt import utils
from jolt import cache
from jolt import log
from jolt import config
from jolt import filesystem as fs
from jolt.tools import Tools
from jolt.error import raise_error_if


NAME = "ftp"
CONNECT_TIMEOUT = 3.5


def catch(func, *args, **kwargs):
    try:
        val = func(*args, **kwargs)
        return val if val is not None else True
    except Exception:
        log.exception()
        return False


class FtpStorage(cache.StorageProvider):
    def __init__(self, cache):
        super(FtpStorage, self).__init__()
        self._cache = cache
        self._uri = config.get(NAME, "host")
        raise_error_if(not self._uri, "ftp URI not configured")
        self._path = config.get(NAME, "path", "")
        self._upload = config.getboolean(NAME, "upload", True)
        self._download = config.getboolean(NAME, "download", True)
        self._tls = config.getboolean(NAME, "tls", False)
        self._disabled = False

    def _get_auth(self):
        service = config.get(NAME, "keyring.service")
        if not service:
            return None, None

        username = config.get(NAME, "keyring.username")
        if not username:
            username = input(NAME + " username: ")
            raise_error_if(not username, "no username configured for " + NAME)
            config.set(NAME, "keyring.username", username)
            config.save()

        password = config.get(NAME, "keyring.password") or keyring.get_password(NAME, username)
        if not password:
            password = getpass.getpass(NAME + " password: ")
            raise_error_if(not password, "no password in keyring for " + NAME)
            keyring.set_password(service, username, password)
        return username, password

    def _get_ftp(self):
        try:
            username, password = self._get_auth()
            if self._tls:
                ftp = FTP_TLS(self._uri, timeout=CONNECT_TIMEOUT)
            else:
                ftp = FTP(self._uri, timeout=CONNECT_TIMEOUT)
            ftp.login(username, password)
            if self._tls:
                ftp.prot_d()
            if not catch(ftp.cwd, self._path):
                if self._path.startswith("/"):
                    ftp.cwd("/")
                components = self._path.split("/")
                for component in components:
                    if not catch(ftp.cwd, component):
                        ftp.mkd(component)
                        ftp.cwd(component)
            return ftp
        except Exception:
            log.exception()
            log.warning("[FTP] failed to establish server connection, disabled")
            self._disabled = True
        return None

    @utils.retried.on_exception((RequestException))
    def download(self, node, force=False):
        if self._disabled:
            return False
        if not self._download and not force:
            return False
        with self._cache.get_artifact(node) as artifact:
            ftp = self._get_ftp()
            if ftp is None:
                return False
            pathname = artifact.get_archive_path()
            name = fs.path.basename(pathname)
            try:
                ftp.cwd(node.canonical_name)
                size = ftp.size(name)
            except Exception:
                return False
            with log.progress("Downloading {0}".format(name), size, "B") as pbar:
                with open(pathname, 'wb') as out_file:
                    def _write(block):
                        out_file.write(block)
                        pbar.update(len(block))
                    return catch(ftp.retrbinary,
                                 "RETR {filename}".format(filename=name),
                                 callback=_write)
        return False

    def download_enabled(self):
        return not self._disabled and self._download

    @utils.retried.on_exception((RequestException))
    def upload(self, node, force=False):
        if self._disabled:
            return True
        if not self._upload and not force:
            return True
        with self._cache.get_artifact(node) as artifact:
            ftp = self._get_ftp()
            if ftp is None:
                return False
            pathname = artifact.get_archive_path()
            name = fs.path.basename(pathname)
            size = Tools().file_size(pathname)
            with log.progress("Uploading {0}".format(name), size, "B") as pbar:
                with open(pathname, 'rb') as in_file:
                    if not catch(ftp.cwd, node.canonical_name):
                        if not catch(ftp.mkd, node.canonical_name):
                            return False
                        if not catch(ftp.cwd, node.canonical_name):
                            return False

                    catch(ftp.delete, name)

                    return catch(ftp.storbinary,
                                 "STOR {filename}".format(filename=name),
                                 in_file,
                                 callback=lambda b: pbar.update(len(b)))
        return False

    def upload_enabled(self):
        return not self._disabled and self._upload

    @utils.retried.on_exception((RequestException))
    def location(self, node):
        if self._disabled:
            return False
        with self._cache.get_artifact(node) as artifact:
            ftp = self._get_ftp()
            if ftp is None:
                return False
            username, _ = self._get_auth()
            username = username + "@" if username is not None else ""
            pathname = artifact.get_archive_path()
            name = fs.path.basename(pathname)
            if not catch(ftp.cwd, node.canonical_name):
                return False
            try:
                if ftp.size(name) is not None:
                    url = "ftp://{user}{uri}/{path}/{taskname}/{archive}".format(
                        user=username,
                        uri=self._uri,
                        path=self._path,
                        taskname=node.canonical_name,
                        archive=name)
                    log.debug("[FTP] {0}", url)
                    return url
            except Exception:
                return False
        return False


@cache.RegisterStorage
class FtpStorageFactory(cache.StorageProviderFactory):
    @staticmethod
    def create(cache):
        log.verbose("[Ftp] Loaded")
        return FtpStorage(cache)



import jolt
from jolt import config
from jolt.error import raise_error_if
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.hooks import TaskHook, TaskHookFactory
from jolt.loader import JoltLoader
import time

try:
    from allure_commons.logger import AllureFileLogger
    from allure_commons.lifecycle import AllureLifecycle
    from allure_commons.model2 import Attachment
    from allure_commons.model2 import Status
    from allure_commons.model2 import StatusDetails
    from allure_commons.model2 import Label
    from allure_commons.types import LabelType
    from allure_commons.utils import host_tag, thread_tag
    from allure_commons.utils import platform_label
    from allure_commons.utils import format_traceback
except ImportError:
    import os
    log.error("Allure plugin enabled but not installed. Install it with: pip install jolt[allure]")
    os._exit(1)

from contextlib import contextmanager
import unittest as ut


__unittest = True


class _ReporterTest(object):
    def __init__(self, reporter, result):
        self._reporter = reporter
        self._task = self._reporter._task
        self._outdirfull = self._reporter._outdirfull
        self._result = result

    def attach(self, name, content, mime_type=None):
        with self._task.tools.cwd(self._outdirfull):
            logpath = utils.sha1(content) + "-" + name
            self._task.tools.write_file(logpath, content, expand=False)
            self._result.attachments.append(
                Attachment(source=logpath, name=name, type=mime_type))

    @contextmanager
    def step(self, desc, *args, **kwargs):
        with self._reporter._lifecycle.start_step() as step:
            step.name = desc

        with log.threadsink() as steplog:
            reporter = _ReporterTest(self._reporter, step)
            try:
                yield reporter
            except ut.SkipTest as e:
                with self._reporter._lifecycle.update_step() as step:
                    step.status = Status.SKIPPED
                    step.statusDetails = StatusDetails(message=str(e))
                    raise e
            except AssertionError as e:
                with self._reporter._lifecycle.update_step() as step:
                    step.status = Status.FAILED
                    step.statusDetails = StatusDetails(message=str(e))
                    raise e
            except Exception as e:
                with self._reporter._lifecycle.update_step() as step:
                    step.status = Status.BROKEN
                    step.statusDetails = StatusDetails(message=str(e))
                    raise e
            else:
                with self._reporter._lifecycle.update_step() as step:
                    step.status = Status.PASSED
            finally:
                # Attach log
                if steplog.getvalue():
                    reporter.attach("log", steplog.getvalue(), "text/plain")
                self._reporter._lifecycle.stop_step()


class Reporter(object):
    def __init__(self, task, builddir=None, results="allure-results"):
        self._host = host_tag()
        self._thread = thread_tag()
        self._task = task
        self._lifecycle = AllureLifecycle()
        self.outdir = builddir or self._task.tools.builddir("allure")
        with self._task.tools.cwd(self.outdir):
            self._outdirfull = self._task.tools.expand_path("allure-results")
        self._logger = AllureFileLogger(self._outdirfull)

    @contextmanager
    def test(self, name, description=None, suite=None):
        with self._lifecycle.schedule_test_case() as result:
            result.name = name
            result.start = time.time() * 1000
            result.fullName = self._task.__class__.__name__ + "." + name
            result.testCaseId = utils.sha1(result.fullName)
            result.historyId = utils.sha1(self._task.identity + result.testCaseId)
            result.description = description
            result.labels.append(Label(name=LabelType.HOST, value=self._host))
            result.labels.append(Label(name=LabelType.THREAD, value=self._thread))
            result.labels.append(Label(name=LabelType.FRAMEWORK, value='jolt'))
            result.labels.append(Label(name=LabelType.LANGUAGE, value=platform_label()))
            if suite:
                result.labels.append(Label(name=LabelType.SUITE, value=suite))
        with log.threadsink() as testlog:
            reporter = _ReporterTest(self, result)
            try:
                yield reporter
            except ut.SkipTest as e:
                result.status = Status.SKIPPED
                result.statusDetails = StatusDetails(message=str(e))
                raise e
            except AssertionError as e:
                with self._lifecycle.update_test_case() as result:
                    result.status = Status.FAILED
                    result.statusDetails = StatusDetails(
                        message=str(e),
                        trace=format_traceback(e.__traceback__))
                    raise e
            except Exception as e:
                with self._lifecycle.update_test_case() as result:
                    result.status = Status.BROKEN
                    result.statusDetails = StatusDetails(
                        message=str(e),
                        trace=format_traceback(e.__traceback__))
                    raise e
            else:
                with self._lifecycle.update_test_case() as result:
                    result.status = Status.PASSED
                    result.statusDetails = StatusDetails()
            finally:
                if testlog.getvalue():
                    reporter.attach("log", testlog.getvalue(), "text/plain")
                with self._lifecycle.update_test_case() as result:
                    result.stop = time.time() * 1000
                self._lifecycle.write_test_case()
                self._logger.report_result(result)

    @contextmanager
    def update_test(self):
        with self._lifecycle.update_test_case() as test:
            yield test

    def publish(self, artifact):
        with self._task.tools.cwd(self.outdir):
            artifact.collect("*")


class Test(jolt.Test):
    def __init__(self, *args, **kwargs):
        super(Test, self).__init__(*args, **kwargs)
        self._curresult = None
        self._host = host_tag()
        self._thread = thread_tag()

    def _run(self, test):
        with self.reporter.test(test.name, test.__doc__, suite=self.short_qualified_name) as result:
            self._curresult = result
            try:
                super()._run(test)
            finally:
                self._curresult = None

    def run(self, deps, tools):
        self.reporter = Reporter(self)
        super().run(deps, tools)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        self.reporter.publish(artifact)

    @contextmanager
    def step(self, name, description=None):
        assert self._curresult, "can't add step when no test is running"
        with self._curresult.step(name, description) as step:
            yield step


class AllureHooks(TaskHook):
    LOGLEVEL = {
        "INFO": log.INFO,
        "VERBOSE": log.VERBOSE,
        "DEBUG": log.DEBUG,
    }

    def __init__(self):
        loglevel = config.get("allure", "loglevel", "INFO")
        self._loglevel = AllureHooks.LOGLEVEL.get(loglevel)
        raise_error_if(not self._loglevel, "allure: illegal loglevel configured: {}", loglevel)
        self._logpath = fs.path.join(JoltLoader.get().joltdir, config.get("allure", "path", "allure-results"))
        fs.rmtree(self._logpath, ignore_errors=True)
        fs.makedirs(self._logpath)
        self._logger = AllureFileLogger(self._logpath)

    def _task_started(self, task):
        task.allure_lifecycle = AllureLifecycle()
        with task.allure_lifecycle.schedule_test_case() as result:
            result.name = task.short_qualified_name
            result.start = time.time() * 1000
            result.fullName = task.qualified_name
            result.description = task.task.__doc__
            result.testCaseId = utils.sha1(result.fullName)
            result.historyId = utils.sha1(task.identity + result.testCaseId)
            result.labels.append(Label(name=LabelType.HOST, value=host_tag()))
            result.labels.append(Label(name=LabelType.THREAD, value=thread_tag()))
            result.labels.append(Label(name=LabelType.FRAMEWORK, value='jolt'))
            result.labels.append(Label(name=LabelType.LANGUAGE, value=platform_label()))
        task.allure_logsink = log.threadsink(self._loglevel)
        task.allure_logsink_buffer = task.allure_logsink.__enter__()

    def _task_ended(self, task, status):
        task.allure_logsink.__exit__(None, None, None)
        with task.allure_lifecycle.update_test_case() as result:
            with task.tools.cwd(self._logpath):
                content = task.allure_logsink_buffer.getvalue()
                if content:
                    logpath = utils.sha1(content) + "-" + "log"
                    task.tools.write_file(logpath, content, expand=False)
                    result.attachments.append(
                        Attachment(source=logpath, name="log", type="text/plain"))
            result.status = status
            if status != Status.SKIPPED:
                result.stop = time.time() * 1000
            else:
                result.start = None
                result.stop = None
            task.allure_lifecycle.write_test_case()
            self._logger.report_result(result)

    def task_started(self, task):
        self._task_started(task)

    def task_failed(self, task):
        self._task_ended(task, Status.FAILED)

    def task_finished(self, task):
        self._task_ended(task, Status.PASSED)

    def task_skipped(self, task):
        self._task_started(task)
        self._task_ended(task, Status.SKIPPED)


@TaskHookFactory.register
class AllureFactory(TaskHookFactory):
    def create(self, env):
        if "allure" in config.plugins():
            return AllureHooks()
        return None



from jolt import attributes as jolt_attributes
from jolt import Download
from jolt import Parameter
from jolt.tasks import TaskRegistry

import platform


@jolt_attributes.system
@jolt_attributes.attribute("bin", "bin_{system}")
@jolt_attributes.attribute("url", "url_{system}")
class NodeJS(Download):
    name = "nodejs"
    version = Parameter("16.15.1", help="NodeJS version.")
    url_linux = "https://nodejs.org/dist/v{version}/node-v{version}-linux-{arch}.tar.gz"
    url_windows = "https://nodejs.org/dist/v{version}/node-v{version}-win-{arch}.zip"
    bin_linux = "node-v{version}-linux-{arch}/bin"
    bin_windows = "node-v{version}-win-{arch}/bin"

    @property
    def arch(self):
        arch = platform.machine()
        if arch == "x86_64":
            return "x64"
        if arch in ["i386", "i486", "i586", "i686"]:
            return "x86"
        if arch in ["arm", "armv7l"]:
            return "armv7l"
        if arch in ["arm64", "armv8l", "aarch64"]:
            return "arm64"
        return None

    def publish(self, artifact, tools):
        super().publish(artifact, tools)
        artifact.environ.PATH.append(self.bin)


TaskRegistry.get().add_task_class(NodeJS)
